"""Symlinking functionality"""
import fnmatch
import itertools
import logging
import os
from pathlib import Path
from typing import Iterable

from . import filesystem as fs
from .gather import load_ender_chest, load_ender_chest_instances, load_shulker_boxes
from .loggers import PLACE_LOGGER
from .prompt import prompt
from .shulker_box import ShulkerBox


def place_ender_chest(
    minecraft_root: Path,
    keep_broken_links: bool = False,
    keep_stale_links: bool = False,
    error_handling: str = "abort",
    relative: bool = True,
    rollback=False,
) -> None:
    """Link all instance files and folders to all shulker boxes

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    keep_broken_links : bool, optional
        By default, this method will remove any broken links in your instances
        and servers folders. To disable this behavior, pass in
        `keep_broken_links=True`.
    keep_stale_links : bool, optional
        By default, this method will remove any links into your EnderChest folder
        that are no longer specified by any shulker box (such as because the
        instance spec or shulker box configuration changed). To disable this
        behavior, pass in `keep_stale_links=True`.
    error_handling : str, optional
        By default, if a linking failure occurs, this method will terminate
        immediately (`error_handling=abort`). Alternatively,
          - pass in `error_handling="ignore"` to continue as if the link failure
            hadn't occurred
          - pass in `error_handling="skip"` to abort linking the current instance
            to the current shulker box but otherwise continue on
          - pass in `error_handling="skip-instance"` to abort linking the current
            instance altogether but to otherwise continue on with other instances
          - pass in `error_handling="skip-shulker-box"` to abort linking to the current
            shulker box altogether but to otherwise continue on with other boxes
          - pass in `error_handling="prompt"` to ask what to do on each failure
    relative : bool, optional
        By default, links will use relative paths when possible. To use absolute
        paths instead (see: https://bugs.mojang.com/projects/MC/issues/MC-263046),
        pass in `relative=False`.
    rollback: bool, optional
        In the future in the event of linking errors passing in `rollback=True`
        can be used to roll back any changes that have already been applied
        based on the error-handling method specified.
    """
    if rollback is not False:
        raise NotImplementedError("Rollbacks are not currently supported")

    try:
        host = load_ender_chest(minecraft_root).name
    except (FileNotFoundError, ValueError) as bad_chest:
        PLACE_LOGGER.error(
            f"Could not load EnderChest from {minecraft_root}:\n  {bad_chest}"
        )
        return

    instances = load_ender_chest_instances(minecraft_root, log_level=logging.DEBUG)

    shulker_boxes: list[ShulkerBox] = []

    for shulker_box in load_shulker_boxes(minecraft_root, log_level=logging.DEBUG):
        if not shulker_box.matches_host(host):
            PLACE_LOGGER.debug(
                f"{shulker_box.name} is not intended for linking to this host ({host})"
            )
            continue
        shulker_boxes.append(shulker_box)

    skip_boxes: list[ShulkerBox] = []

    def handle_error(shulker_box: ShulkerBox | None) -> str:
        """Centralized error-handling

        Parameters
        ----------
        shulker_box:
            The current shulker box (in case it needs to be added to the skip list)

        Returns
        -------
        str
            Instructions on what to do next. Options are:
              - return
              - break
              - continue
              - pass
        """
        if error_handling == "prompt":
            proceed_how = (
                prompt(
                    "How would you like to proceed?"
                    "\n[Q]uit, [C]ontinue, abort linking the rest of this shulker/instance [M]atch?"
                    "\nskip the rest of this [I]nstance, skip the rest of this [S]hulker box?",
                    suggestion="I",  # TODO: reevaluate default (it's 50/50 due to link folders)
                )
                .lower()
                .replace(" ", "")
                .replace("-", "")
                .replace("_", "")
            )
            match proceed_how:
                case "" | "i" | "instance" | "skipinstance":
                    proceed_how = "skip-instance"
                case "q" | "quit" | "abort" | "exit" | "stop":
                    proceed_how = "abort"
                case "c" | "continue" | "ignore":
                    proceed_how = "ignore"
                case "m" | "match" | "skip":
                    proceed_how = "skip"
                case "s" | "shulker" | "shulkerbox" | "skipshulker":
                    proceed_how = "skip-shulker"
                case _:
                    PLACE_LOGGER.error("Invalid selection.")
                    return handle_error(shulker_box)
        else:
            proceed_how = error_handling

        match proceed_how:
            case "abort" | "stop" | "quit" | "exit":
                PLACE_LOGGER.error("Aborting")
                return "return"
            case "ignore":
                PLACE_LOGGER.debug("Ignoring")
                return "pass"
            case "skip":
                PLACE_LOGGER.warning("Skipping the rest of this match")
                return "continue"
            case "skip-instance":
                PLACE_LOGGER.warning("Skipping any more linking from this instance")

                return "break"
            case "skip-shulker-box" | "skip-shulkerbox" | "skip-shulker":
                PLACE_LOGGER.warning("Skipping any more linking into this shulker box")
                if shulker_box:
                    skip_boxes.append(shulker_box)
                return "continue"
            case _:
                raise ValueError(
                    f"Unrecognized error-handling method: {error_handling}"
                )

    for instance in instances:
        instance_root = (minecraft_root / instance.root.expanduser()).expanduser()
        if not instance_root.exists():
            PLACE_LOGGER.error(
                "No minecraft instance exists at"
                f" {instance_root.expanduser().absolute()}"
            )
            match handle_error(None):
                case "return":
                    return
                case "break":
                    break
                case _:  # nothing to link, so might as well skip the rest
                    continue

        # start by removing all existing symlinks into the EnderChest
        if not keep_stale_links:
            for file in instance_root.rglob("*"):
                if file.is_symlink():
                    if fs.links_into_enderchest(minecraft_root, file):
                        PLACE_LOGGER.debug(
                            f"Removing old link: {file} -> {os.readlink(file)}"
                        )
                        file.unlink()

        for shulker_box in shulker_boxes:
            if not shulker_box.matches(instance):
                continue
            if shulker_box in skip_boxes:
                continue

            box_root = shulker_box.root.expanduser().absolute()

            PLACE_LOGGER.info(f"Linking {instance.root} to {shulker_box.name}")

            resources = set(_rglob(box_root, shulker_box.max_link_depth))

            match_exit = "pass"
            for link_folder in shulker_box.link_folders:
                resources -= {box_root / link_folder}
                resources -= set((box_root / link_folder).rglob("*"))
                try:
                    link_resource(link_folder, box_root, instance_root, relative)
                except (OSError, NotADirectoryError) as oh_no:
                    PLACE_LOGGER.error(
                        f"Error linking shulker box {shulker_box.name}"
                        f" to instance {instance.name}:"
                        f"\n  {(instance.root / link_folder)} is a"
                        " non-empty directory"
                    )
                    match handle_error(shulker_box):
                        case "return":
                            return
                        case "break":
                            match_exit = "break"
                            break
                        case "continue":
                            match_exit = "continue"
                            break
                        case "pass":
                            continue  # or pass--it's the end of the loop

            if match_exit not in ("break", "continue"):
                for resource in resources:
                    resource_path = resource.relative_to(box_root)
                    for pattern in shulker_box.do_not_link:
                        if fnmatch.fnmatchcase(
                            str(resource_path), pattern
                        ) or fnmatch.fnmatchcase(
                            str(resource_path), os.path.join("*", pattern)
                        ):
                            PLACE_LOGGER.debug(
                                "Skipping %s (matches pattern %s)",
                                resource_path,
                                pattern,
                            )
                            break
                    else:
                        try:
                            link_resource(
                                resource_path,
                                box_root,
                                instance_root,
                                relative,
                            )
                        except (OSError, NotADirectoryError) as oh_no:
                            PLACE_LOGGER.error(
                                f"Error linking shulker box {shulker_box.name}"
                                f" to instance {instance.name}:"
                                f"\n  {(instance.root / resource_path)}"
                                " already exists"
                            )
                            match handle_error(shulker_box):
                                case "return":
                                    return
                                case "break":
                                    match_exit = "break"
                                    break
                                case "continue":
                                    match_exit = "continue"  # technically does nothing
                                    break
                                case "pass":
                                    continue  # or pass--it's the end of the loop

            # consider this a "finally"
            if not keep_broken_links:
                # we clean up as we go, just in case of a failure
                for file in instance_root.rglob("*"):
                    if not file.exists():
                        PLACE_LOGGER.debug(f"Removing broken link: {file}")
                        file.unlink()

            if match_exit == "break":
                break


def link_resource(
    resource_path: str | Path,
    shulker_root: Path,
    instance_root: Path,
    relative: bool,
) -> None:
    """Create a symlink for the specified resource from an instance's space
    pointing to the tagged file / folder living inside a shulker box.

    Parameters
    ----------
    resource_path : str or Path
        Location of the resource relative to the instance's ".minecraft" folder
    shulker_root : Path
        The path to the shulker box
    instance_root : Path
        The path to the instance's ".minecraft" folder
    relative : bool
        If True, the link will be use a relative path if possible. Otherwise,
        an absolute path will be used, regardless of whether a a relative or
        absolute path was provided.

    Raises
    ------
    NotADirectoryError
        If a file already exists where you're attempting to place the symlink
    OSError
        If a non-empty directory already exists where you're attempting to
        place the symlink

    Notes
    -----
    - This method will create any folders that do not exist within an instance
    - This method will overwrite existing symlinks and empty folders
      but will not overwrite or delete any actual files.
    """
    instance_path = (instance_root / resource_path).expanduser().absolute()
    instance_path.parent.mkdir(parents=True, exist_ok=True)

    target: str | Path = (shulker_root / resource_path).expanduser().absolute()
    if relative:
        target = os.path.relpath(target, instance_path.parent)

    if instance_path.is_symlink():
        # remove previous symlink in this spot
        instance_path.unlink()
        PLACE_LOGGER.debug(f"Removed previous link at {instance_path}")
    else:
        try:
            os.rmdir(instance_path)
            PLACE_LOGGER.debug(f"Removed empty directory at {instance_path}")
        except FileNotFoundError:
            pass  # A-OK

    PLACE_LOGGER.debug(f"Linking {instance_path} to {target}")
    os.symlink(
        target,
        instance_path,
        target_is_directory=(shulker_root / resource_path).is_dir(),
    )


def _rglob(root: Path, max_depth: int) -> Iterable[Path]:
    """Find all files (and directories* and symlinks) in the path up to the
    specified depth

    Parameters
    ----------
    root : Path
        The path to search
    max_depth : int
        The maximum number of levels to go

    Returns
    -------
    list-like of paths
        The files (and directories and symlinks) in the path up to that depth

    Notes
    -----
    - Unlike an actual rglob, this method does not return any directories that
      are not at the maximum depth
    - Setting max_depth to 0 (or below) will return all files in the root, but
      ***be warned*** that because this method follows symlinks, you can very
      easily find yourself in an infinite loop
    """
    top_level = root.iterdir()
    if max_depth == 1:
        return top_level
    return itertools.chain(
        *(
            _rglob(path, max_depth - 1) if path.is_dir() else (path,)
            for path in top_level
        )
    )



"""Tests around file discovery and registration"""
import logging
import os
import re
from pathlib import Path

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import gather
from enderchest import instance as i

from . import utils


class TestListShulkerBoxes:
    @pytest.fixture(autouse=True)
    def populate_shulker_boxes(self, minecraft_root):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

        # also write a bad shulker box
        bad_ini = minecraft_root / "EnderChest" / "not_ini" / fs.SHULKER_BOX_CONFIG_NAME
        bad_ini.parent.mkdir()
        bad_ini.write_text("is_this_valid_ini=no")

    def test_list_shulker_box_reports_the_boxes_in_order(self, minecraft_root, caplog):
        _ = gather.load_shulker_boxes(minecraft_root)
        assert (
            """0. global
  1. 1.19
  2. vanilla
  3. optifine"""
            in caplog.records[-1].msg % caplog.records[-1].args
        )

    def test_list_shulker_box_doesnt_choke_on_relative_root(
        self, minecraft_root, caplog, monkeypatch
    ):
        monkeypatch.chdir(minecraft_root.parent)
        _ = gather.load_shulker_boxes(Path(minecraft_root.name))
        assert (
            """0. global
  1. 1.19
  2. vanilla
  3. optifine"""
            in caplog.records[-1].msg % caplog.records[-1].args
        )

    def test_list_shulker_box_warns_if_theres_a_bad_box(self, minecraft_root, caplog):
        _ = gather.load_shulker_boxes(minecraft_root)

        warnings = [
            record for record in caplog.records if record.levelname == "WARNING"
        ]

        assert re.search(
            rf"Could not parse(.*)not_ini(.*){fs.SHULKER_BOX_CONFIG_NAME}",
            warnings[-1].msg % warnings[-1].args,
        )


class TestLoadBoxInstanceMatches:
    @pytest.fixture(autouse=True)
    def populate_shulker_boxes(self, minecraft_root, home):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

    @pytest.mark.parametrize(
        "instance_name",
        [mc.name for mc in utils.TESTING_INSTANCES] + ["unown"],
    )
    def test_loading_boxes_that_match_instance(self, minecraft_root, instance_name):
        # because TESTING_SHULKER_INSTANCE_MATCHES uses the path
        instance_name_lookup = {
            "official": "~",
            "Chest Boat": "chest-boat",
        }
        box_lookup = {
            box.name: box
            for box in gather.load_shulker_boxes(
                minecraft_root, log_level=logging.DEBUG
            )
        }

        expected = []
        for box_name, mc_name, should_match in utils.TESTING_SHULKER_INSTANCE_MATCHES:
            if mc_name == instance_name_lookup.get(instance_name, instance_name):
                if should_match:
                    expected.append(box_lookup[box_name])

        assert (
            gather.get_shulker_boxes_matching_instance(minecraft_root, instance_name)
            == expected
        )

    @pytest.mark.parametrize(
        "shulker_box_name",
        list({box for _, box, _ in utils.TESTING_SHULKER_INSTANCE_MATCHES}) + ["unown"],
    )
    def test_loading_instances_that_match_boxes(self, minecraft_root, shulker_box_name):
        instance_lookup = {
            mc.name: mc
            for mc in gather.load_ender_chest_instances(
                minecraft_root, log_level=logging.DEBUG
            )
        }
        # because TESTING_SHULKER_INSTANCE_MATCHES uses the path
        instance_lookup["~"] = instance_lookup["official"]
        instance_lookup["chest-boat"] = instance_lookup["Chest Boat"]

        expected = []
        for (
            box_name,
            instance_name,
            should_match,
        ) in utils.TESTING_SHULKER_INSTANCE_MATCHES:
            if box_name == shulker_box_name:
                if should_match:
                    expected.append(instance_lookup[instance_name])

        assert (
            gather.get_instances_matching_shulker_box(minecraft_root, shulker_box_name)
            == expected
        )


class TestGatherInstances:
    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root, home):
        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False)
        )
        with fs.ender_chest_config(minecraft_root).open("w") as ec_config:
            ec_config.write(
                """
[properties]
offer-to-update-symlink-allowlist = False
"""
            )

        yield

        # tests that with offer-to-update set to False, EnderChest doesn't go rogue
        assert not (home / ".minecraft" / "allowed_symlinks.txt").exists()

    def test_official_instance_parsing(self, home):
        assert utils.normalize_instance(
            gather.gather_metadata_for_official_instance(home / ".minecraft")
        ) == utils.normalize_instance(utils.TESTING_INSTANCES[0])

    def test_instance_search_finds_official_instance(self, minecraft_root, home):
        assert i.equals(
            minecraft_root,
            gather.gather_minecraft_instances(minecraft_root, home, official=True)[0],
            utils.TESTING_INSTANCES[0],
        )

    @pytest.mark.parametrize(
        "instance, idx",
        (
            ("axolotl", 1),
            ("bee", 2),
            ("chest-boat", 3),
        ),
    )
    def test_mmc_instance_parsing(self, minecraft_root, instance, idx):
        assert utils.normalize_instance(
            gather.gather_metadata_for_mmc_instance(
                minecraft_root / "instances" / instance / ".minecraft"
            )
        ) == utils.normalize_instance(
            # we're not testing aliasing right now
            utils.TESTING_INSTANCES[idx]._replace(name=instance)
        )

    def test_instance_search_finds_mmc_instances(self, minecraft_root):
        instances = sorted(
            gather.gather_minecraft_instances(
                minecraft_root, minecraft_root, official=False
            ),
            key=lambda instance: instance.name,
        )

        assert len(instances) == 3

        assert all(
            [
                i.equals(
                    minecraft_root, instances[idx - 1], utils.TESTING_INSTANCES[idx]
                )
                for idx in range(1, 4)
            ]
        )

    def test_instance_search_doesnt_choke_on_relative_root(
        self, minecraft_root, monkeypatch
    ):
        monkeypatch.chdir(minecraft_root.parent)
        instances = sorted(
            gather.gather_minecraft_instances(
                Path(minecraft_root.name), Path(minecraft_root.name), official=False
            ),
            key=lambda instance: instance.name,
        )

        assert len(instances) == 3

        assert all(
            [
                i.equals(
                    minecraft_root, instances[idx - 1], utils.TESTING_INSTANCES[idx]
                )
                for idx in range(1, 4)
            ]
        )

    def test_instance_search_can_find_all_instances(self, minecraft_root, home):
        instances = sorted(
            gather.gather_minecraft_instances(
                minecraft_root, minecraft_root, official=None
            ),
            key=lambda instance: instance.name
            if instance.name != "official"
            else "aaa",  # sorting hack
        )

        assert len(instances) == 4

        assert all(
            [
                i.equals(minecraft_root, instances[idx], utils.TESTING_INSTANCES[idx])
                for idx in range(4)
            ]
        )

    def test_instance_search_warns_if_no_instances_can_be_found(
        self, minecraft_root, caplog
    ):
        empty_folder = minecraft_root / "nothing in there"
        empty_folder.mkdir()
        instances = gather.gather_minecraft_instances(
            minecraft_root, empty_folder, official=None
        )

        assert len(instances) == 0

        warn_log = "\n".join(
            record.msg for record in caplog.records if record.levelname == "WARNING"
        )

        assert "Could not find any Minecraft instances" in warn_log

    def test_onboarding_new_instances(self, minecraft_root, home):
        # start with a blank chest
        craft.craft_ender_chest(minecraft_root, remotes=())
        gather.update_ender_chest(minecraft_root, ("~", minecraft_root))

        instances = sorted(
            gather.load_ender_chest_instances(minecraft_root),
            key=lambda instance: instance.name
            if instance.name != "official"
            else "aaa",  # sorting hack
        )

        assert len(instances) == 4

        assert all(
            [
                i.equals(minecraft_root, instances[idx], utils.TESTING_INSTANCES[idx])
                for idx in range(4)
            ]
        )


class TestSymlinkAllowlistVersionChecker:
    """aka check my regex"""

    @pytest.mark.parametrize(
        "version_string, expected",
        (
            ("1.19.4", False),
            ("1.20", True),
            ("1.20.1-pre1", True),
            ("1.21-pre1", True),
            ("1.19-rc2", False),
            ("1.2.1", False),
            ("22w43a", False),
            ("23w9a", False),
            ("23w09a", False),
            ("23w51a", True),
            ("24w6a", True),
            ("24w06a", True),
        ),
    )
    def test_check_if_version_needs_allow_list(self, version_string, expected):
        assert gather._needs_symlink_allowlist(version_string) == expected


class TestSymlinkAllowlistHandling:
    @pytest.fixture(autouse=True)
    def setup_teardown(self, minecraft_root):
        utils.pre_populate_enderchest(
            fs.ender_chest_folder(minecraft_root, check_exists=False)
        )
        with fs.ender_chest_config(minecraft_root).open("w") as ec_config:
            ec_config.write(
                """
[properties]
"""
            )

        utils.populate_mmc_instance_folder(
            minecraft_root / "instances" / "talespin", "1.20", "Quilt", "talespin"
        )

        (
            minecraft_root
            / "instances"
            / "talespin"
            / ".minecraft"
            / "allowed_symlinks.txt"
        ).write_text(
            "my_development_folder\n"
        )  # read: not EnderChest

        yield

        assert (
            (
                minecraft_root
                / "instances"
                / "talespin"
                / ".minecraft"
                / "allowed_symlinks.txt"
            )
            .read_text()
            .startswith("my_development_folder\n")
        )

    @pytest.mark.parametrize("instance", ("official", "talespin"))
    def test_ender_chest_does_not_write_allowlist_without_consent(
        self, minecraft_root, home, monkeypatch, capsys, instance
    ):
        if instance == "official":
            search_path = home
        else:
            search_path = minecraft_root / "instances" / "talespin"
        i_forbid_it = utils.scripted_prompt(["no"])
        monkeypatch.setattr("builtins.input", i_forbid_it)

        gather.gather_minecraft_instances(minecraft_root, search_path, None)

        _ = capsys.readouterr()  # suppress outputs

        # easier to check both
        assert not (home / ".minecraft" / "allowed_symlinks.txt").exists()
        assert (
            minecraft_root
            / "instances"
            / "talespin"
            / ".minecraft"
            / "allowed_symlinks.txt"
        ).read_text() == "my_development_folder\n"

        # make sure all responses were used
        with pytest.raises(StopIteration):
            i_forbid_it()

    def test_ender_chest_will_write_allowlists_with_consent(
        self, minecraft_root, home, monkeypatch, capsys
    ):
        mkay = utils.scripted_prompt(["y"] * 2)
        monkeypatch.setattr("builtins.input", mkay)

        gather.gather_minecraft_instances(minecraft_root, home, True)

        # this is also testing that you're not getting prompted
        # for pre-1.20 instances
        gather.gather_minecraft_instances(
            minecraft_root, minecraft_root / "instances", False
        )

        _ = capsys.readouterr()  # suppress outputs

        ender_chest_path = os.path.abspath(fs.ender_chest_folder(minecraft_root))

        assert (
            home / ".minecraft" / "allowed_symlinks.txt"
        ).read_text() == ender_chest_path + "\n"
        assert (
            minecraft_root
            / "instances"
            / "talespin"
            / ".minecraft"
            / "allowed_symlinks.txt"
        ).read_text() == f"my_development_folder\n{ender_chest_path}\n"

        # make sure all responses were used
        with pytest.raises(StopIteration):
            mkay()



"""Test the command-line interface"""
import logging
import os
from pathlib import Path

import pytest

import enderchest
from enderchest import cli, place, remote


class TestHelp:
    @pytest.mark.parametrize("help_flag", ("-h", "--help"))
    def test_help_displays_version(self, capsys, help_flag):
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", help_flag])

        assert enderchest.__version__ in capsys.readouterr().out

    @pytest.mark.parametrize("help_flag", ("-h", "--help"))
    def test_help_ignores_arguments_that_follow(self, capsys, help_flag):
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", help_flag, "foo"])

        assert "foo" not in capsys.readouterr().out
        assert "foo" not in capsys.readouterr().err


class TestVersion:
    @pytest.mark.parametrize("version_flag", ("-v", "--version"))
    def test_version_displays_version(self, capsys, version_flag):
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", version_flag])

        assert enderchest.__version__ in capsys.readouterr().out

    @pytest.mark.parametrize("version_flag", ("-v", "--version"))
    def test_help_ignores_arguments_that_follow(self, capsys, version_flag):
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", version_flag, "foo"])

        assert "foo" not in capsys.readouterr().out
        assert "foo" not in capsys.readouterr().err


class ActionTestSuite:
    required_args: tuple[str, ...] = ()

    @pytest.mark.parametrize("help_flag", ("-h", "--help"))
    def test_help_gives_action_specific_help(
        self,
        capsys,
        help_flag,
    ):
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", *self.action.split(), help_flag])

        stdout = capsys.readouterr().out
        assert f"enderchest {self.action} [-h]" in stdout

    def test_default_root_is_cwd(self, monkeypatch):
        monkeypatch.setattr(os, "getcwd", lambda: "~~dummy~~")
        _, root, _, _ = cli.parse_args(
            ["enderchest", *self.action.split(), *self.required_args]
        )
        assert root == Path("~~dummy~~")

    def test_first_argument_is_root(self):
        _, root, _, _ = cli.parse_args(
            ["enderchest", *self.action.split(), "/home", *self.required_args]
        )
        assert root == Path("/home")

    def test_root_can_also_be_provided_by_flag(self):
        _, root, _, _ = cli.parse_args(
            ["enderchest", *self.action.split(), *self.required_args, "--root", "/home"]
        )
        assert root == Path("/home")

    @pytest.mark.parametrize(
        "verbosity_flag, expected_verbosity",
        (
            ("-v", logging.DEBUG - 9),
            ("-q", logging.WARNING - 9),
            ("--verbose", logging.DEBUG - 9),
            ("--quiet", logging.WARNING - 9),
            ("-vv", -9),
            ("-qq", logging.ERROR - 9),
            ("-vvqvqqvqv", logging.DEBUG - 9),
        ),
    )
    def test_altering_verbosity(self, verbosity_flag, expected_verbosity):
        _, _, log_level, _ = cli.parse_args(
            ["enderchest", *self.action.split(), verbosity_flag, *self.required_args]
        )

        assert log_level == expected_verbosity


class TestCraft(ActionTestSuite):
    action = "craft"

    @pytest.mark.parametrize("remote_flag", ("-r", "--remote"))
    def test_passing_in_a_single_remote(self, remote_flag):
        *_, options = cli.parse_args(
            [
                "enderchest",
                "craft",
                remote_flag,
                "sftp://openbagtwo@mirror/home/openbagtwo/minecraft",
            ]
        )

        assert options["remotes"] == [
            "sftp://openbagtwo@mirror/home/openbagtwo/minecraft"
        ]

    @pytest.mark.parametrize("remote_flag_1", ("-r", "--remote"))
    @pytest.mark.parametrize("remote_flag_2", ("-r", "--remote"))
    @pytest.mark.parametrize("remote_flag_3", ("-r", "--remote"))
    def test_passing_in_a_multiple_remotes_plus_other_kwargs(
        self, remote_flag_1, remote_flag_2, remote_flag_3
    ):
        *_, options = cli.parse_args(
            [
                "enderchest",
                "craft",
                remote_flag_1,
                "sftp://openbagtwo@mirror/home/openbagtwo/minecraft",
                remote_flag_2,
                "file://~/minecraft2",
                "--overwrite",
                remote_flag_3,
                "sftp://pie/opt/run/minecraft",
            ]
        )

        assert options["remotes"] == [
            "sftp://openbagtwo@mirror/home/openbagtwo/minecraft",
            "file://~/minecraft2",
            "sftp://pie/opt/run/minecraft",
        ]


class TestCraftShulker(ActionTestSuite):
    action = "craft shulker_box"
    required_args = ("nombre",)

    @pytest.mark.parametrize("instance_flag", ("-i", "--instance"))
    def test_passing_in_a_single_instance(self, instance_flag):
        *_, options = cli.parse_args(
            ["enderchest", "craft", "shulker_box", instance_flag, "endcity", "spitty"]
        )

        # remove unset stuff
        options = {key: value for key, value in options.items() if value}
        assert options == {
            "instances": ["endcity"],
            "name": "spitty",
        }

    def test_passing_in_a_mix_of_multi_kwargs(
        self,
    ):
        _, root, _, options = cli.parse_args(
            [
                "enderchest",
                "craft",
                "shulker_box",
                "--overwrite",
                "--instance",
                "onion",
                "rooty",
                "-t",
                "allium*" "rutabaga",
                "--instance",
                "turnip",
                "--enderchest",
                "farm",
            ]
        )

        # remove unset stuff
        options = {key: value for key, value in options.items() if value}
        assert root, options == (
            "rooty",
            {
                "name": "rutabaga",
                "instances": ["onion", "turnip"],
                "tags": ["allium"],
                "hosts": ["farm"],
                "overwrite": True,
            },
        )


class TestPlace(ActionTestSuite):
    action = "place"

    @pytest.fixture
    def place_log(self, monkeypatch):
        place_log: list[tuple[Path, dict]] = []

        def mock_place(path, **kwargs):
            place_log.append((path, kwargs))

        monkeypatch.setattr(place, "place_ender_chest", mock_place)

        yield place_log

    @pytest.mark.parametrize("option", ("keep_broken_links", "keep_stale_links"))
    def test_remove_broken_links_by_default(self, option, place_log):
        action, minecraft_root, _, options = cli.parse_args(
            ["enderchest", "place", "/home"]
        )
        action(minecraft_root, **options)
        assert len(place_log) == 1
        assert place_log[0][1][option] is False

    @pytest.mark.parametrize("flag", ("-k", "--keep-stale-links"))
    def test_keep_stale_links(self, flag, place_log):
        action, minecraft_root, _, options = cli.parse_args(
            ["enderchest", "place", "/home", flag]
        )
        action(minecraft_root, **options)
        assert len(place_log) == 1
        assert place_log[0][1]["keep_stale_links"] is True

    @pytest.mark.parametrize("flag", ("-kk", "--keep-broken-links"))
    def test_keep_broken_links(self, flag, place_log):
        action, minecraft_root, _, options = cli.parse_args(
            ["enderchest", "place", "/home", flag]
        )
        action(minecraft_root, **options)
        assert len(place_log) == 1
        assert place_log[0][1]["keep_broken_links"] is True

    def test_prompt_on_error_by_default(self):
        *_, options = cli.parse_args(["enderchest", "place"])
        assert (
            options["errors"],
            options["stop_at_first_failure"],
            options["ignore_errors"],
        ) == ("prompt", False, False)

    @pytest.mark.parametrize(
        "flags",
        (("-x",), ("--stop-at-first-failure",), ("--errors", "abort")),
        ids=("-x", "--stop-at-first-failure", "--errors=abort"),
    )
    def test_stop_on_error(self, place_log, flags):
        action, *_, options = cli.parse_args(["enderchest", "place", *flags])
        action(Path(), **options)
        assert place_log[0][1]["error_handling"] == "abort"

    @pytest.mark.parametrize(
        "flags",
        (("--ignore-errors",), ("--errors", "ignore")),
        ids=("--ignore-errors", "--errors=ignore"),
    )
    def test_ignore_errors(self, place_log, flags):
        action, *_, options = cli.parse_args(["enderchest", "place", *flags])
        action(Path(), **options)
        assert place_log[0][1]["error_handling"] == "ignore"

    def test_absolute_path_links_is_the_default(self, place_log):
        action, *_, options = cli.parse_args(["enderchest", "place"])
        action(Path(), **options)
        assert place_log[0][1]["relative"] is False

    @pytest.mark.parametrize(
        "flag, expected",
        (("-a", False), ("--absolute", False), ("-r", True), ("--relative", True)),
    )
    def test_explicitly_specify_abs_or_rel(self, place_log, flag, expected):
        action, *_, options = cli.parse_args(["enderchest", "place"])
        action(Path(), **options)
        assert place_log[0][1]["relative"] is False


class TestGather(ActionTestSuite):
    action = "gather minecraft"
    required_args = ("~",)

    @pytest.mark.parametrize("with_root", (False, True), ids=("no_root", "with-root"))
    def test_gather_requires_at_least_one_search_path(self, with_root, capsys):
        more_args = ("--root", "/minecraft") if with_root else ()
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", *self.action.split(), *more_args])

        _ = capsys.readouterr()  # suppress outputs

    @pytest.mark.parametrize("with_root", (False, True), ids=("no_root", "with-root"))
    def test_single_arg_interpreted_as_search_path(self, with_root, capsys):
        more_args = ("--root", ".") if with_root else ()

        _, root, _, options = cli.parse_args(
            ["enderchest", *self.action.split(), *more_args, "~"]
        )

        _ = capsys.readouterr()  # suppress outputs

        assert (root.resolve(), options["search_paths"]) == (
            Path(".").resolve(),
            [Path("~")],
        )

    def test_first_arg_interpreted_as_root(self):  # I actually really don't like this
        _, root, _, options = cli.parse_args(
            [
                "enderchest",
                *self.action.split(),
                ".",
                "~",
                "/here",
                "/there",
                "everywhere",
            ]
        )
        assert (root.resolve(), options["search_paths"]) == (
            Path(".").resolve(),
            [Path("~"), Path("/here"), Path("/there"), Path("everywhere")],
        )


class TestGatherRemote(ActionTestSuite):
    action = "gather enderchests"
    required_args = ("sftp://openbagtwo@steamdeck/home/deck",)

    @pytest.mark.parametrize("with_root", (False, True), ids=("no_root", "with-root"))
    def test_gather_requires_at_least_one_remote(self, with_root, capsys):
        more_args = ("--root", "/minecraft") if with_root else ()
        with pytest.raises(SystemExit):
            cli.parse_args(["enderchest", *self.action.split(), *more_args])

        _ = capsys.readouterr()  # suppress outputs

    @pytest.mark.parametrize("with_root", (False, True), ids=("no_root", "with-root"))
    def test_single_arg_interpreted_as_remote(self, with_root):
        more_args = ("--root", ".") if with_root else ()

        _, root, _, options = cli.parse_args(
            [
                "enderchest",
                *self.action.split(),
                *more_args,
                "sftp://openbagtwo@steamdeck:~",
            ]
        )
        assert (root.resolve(), options["remotes"]) == (
            Path(".").resolve(),
            ["sftp://openbagtwo@steamdeck:~"],
        )

    def test_first_arg_interpreted_as_root(self):  # I actually really don't like this
        _, root, _, options = cli.parse_args(
            [
                "enderchest",
                *self.action.split(),
                "~",
                "sftp://openbagtwo@steamdeck/home/deck",
                "ipoac://birdhouse/your/soul",
                "sneakernet://123.fake.street/mailbox",
            ]
        )
        assert (root.expanduser().resolve(), options["remotes"]) == (
            Path("~").expanduser().resolve(),
            [
                "sftp://openbagtwo@steamdeck/home/deck",
                "ipoac://birdhouse/your/soul",
                "sneakernet://123.fake.street/mailbox",
            ],
        )


class TestInstanceInventory(ActionTestSuite):
    action = "inventory minecraft"
    required_args = ("cherry grove",)


class TestShulkerInventory(ActionTestSuite):
    action = "inventory shulker_box"
    required_args = ("nombre",)


class TestOpen:
    action = "open"
    op = "pull"

    def test_op_is_routed_successfully(self, monkeypatch):
        sync_log: list[tuple[str, str, dict]] = []

        def mock_sync(root, op, **kwargs):
            sync_log.append((root, op, kwargs))

        monkeypatch.setattr(remote, "sync_with_remotes", mock_sync)

        action, root, _, kwargs = cli.parse_args(["enderchest", *self.action.split()])
        action(root, **kwargs)

        assert len(sync_log) == 1
        assert sync_log[0][1] == self.op

    @pytest.mark.parametrize(
        "verbosity_flag, expected_verbosity",
        (
            ("-v", 1),
            ("-q", -1),
            ("--verbose", 1),
            ("--quiet", -1),
            ("-vv", 2),
            ("-qq", -2),
            ("-vvqvqqvqv", 1),
        ),
    )
    def test_verbosity_modifier_is_passed_to_op(
        self, monkeypatch, verbosity_flag, expected_verbosity
    ):
        sync_log: list[tuple[str, str, dict]] = []

        def mock_sync(root, op, **kwargs):
            sync_log.append((root, op, kwargs))

        monkeypatch.setattr(remote, "sync_with_remotes", mock_sync)

        action, root, _, kwargs = cli.parse_args(
            ["enderchest", *self.action.split(), verbosity_flag]
        )
        action(root, **kwargs)

        assert len(sync_log) == 1
        assert sync_log[0][2]["verbosity"] == expected_verbosity

    def test_dry_run_is_false_by_default(self, monkeypatch):
        sync_log: list[tuple[str, str, dict]] = []

        def mock_sync(root, op, **kwargs):
            sync_log.append((root, op, kwargs))

        monkeypatch.setattr(remote, "sync_with_remotes", mock_sync)

        action, root, _, kwargs = cli.parse_args(["enderchest", *self.action.split()])
        action(root, **kwargs)

        assert len(sync_log) == 1
        assert sync_log[0][2]["dry_run"] is False

    def test_passing_a_bunch_of_args(self, monkeypatch):
        sync_log: list[tuple[str, str, dict]] = []

        def mock_sync(root, op, **kwargs):
            sync_log.append((root, op, kwargs))

        monkeypatch.setattr(remote, "sync_with_remotes", mock_sync)

        action, root, _, kwargs = cli.parse_args(
            [
                "enderchest",
                *self.action.split(),
                "--root",
                ".",
                "--dry-run",
                "-t",
                "15",
                "-w",
                "0",
                "-e",
                "private",
                "*.secret",
            ]
        )
        action(root, **kwargs)

        assert len(sync_log) == 1
        assert sync_log[0][2] == {
            "dry_run": True,
            "timeout": 15,
            "sync_confirm_wait": 0,
            "exclude": ["private", "*.secret"],
            "verbosity": 0,
        }


class TestClose(TestOpen):
    action = "close"
    op = "push"



"""Low-level functionality for synchronizing across different machines"""
import importlib
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator, Iterable
from urllib.parse import ParseResult

from ..loggers import SYNC_LOGGER
from .utils import get_default_netloc, path_from_uri, render_remote

SUPPORTED_PROTOCOLS = ("rsync", "file")

DEFAULT_PROTOCOL = SUPPORTED_PROTOCOLS[0]


def pull(
    remote_uri: ParseResult,
    local_path: Path,
    exclude: Iterable[str] | None = None,
    dry_run: bool = False,
    **kwargs,
) -> None:
    """Pull all upstream changes from a remote into the specified location

    Parameters
    ----------
    remote_uri : ParseResult
        The URI for the remote resource to pull
    local_path : Path
        The local destination
    exclude : list of str, optional
        Any patterns that should be excluded from the sync
    dry_run : bool, optional
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them). Default is False.
    **kwargs
        Any additional options to pass into the sync command
    """
    try:
        protocol = importlib.import_module(f"{__package__}.{remote_uri.scheme.lower()}")
        protocol.pull(remote_uri, local_path, exclude or (), dry_run, **kwargs)
    except ModuleNotFoundError:  # pragma: no cover
        raise NotImplementedError(
            f"Protocol {remote_uri.scheme} is not currently implemented"
        )
    except TypeError as unknown_kwarg:
        raise NotImplementedError(
            f"Protocol {remote_uri.scheme} does not support that functionality:\n  {unknown_kwarg}"
        )


def push(
    local_path: Path,
    remote_uri: ParseResult,
    exclude: Iterable[str] | None = None,
    dry_run: bool = False,
    **kwargs,
) -> None:
    """Push all local changes in the specified directory into the specified remote

    Parameters
    ----------
    local_path
        The local path to push
    remote_uri : ParseResult
        The URI for the remote destination
    exclude : list of str, optional
        Any patterns that should be excluded from the sync
    dry_run : bool, optional
        Whether to only simulate this sync (report the operations to be performed
        but not actually perform them). Default is False.
    **kwargs
        Any additional options to pass into the sync command
    """
    try:
        protocol = importlib.import_module(f"{__package__}.{remote_uri.scheme.lower()}")
        protocol.push(local_path, remote_uri, exclude or (), dry_run, **kwargs)
    except ModuleNotFoundError:
        raise NotImplementedError(
            f"Protocol {remote_uri.scheme} is not currently implemented"
        )


@contextmanager
def remote_file(uri: ParseResult) -> Generator[Path, None, None]:
    """Grab a file from a remote filesystem by its URI and read its contents

    Parameters
    ----------
    uri : parsed URI
        The URI of the file to read

    Yields
    ------
    Path
        A path to a local (temp) copy of the file
    """
    with TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
        pull(uri, Path(tmpdir), verbosity=-1)
        yield Path(tmpdir) / Path(uri.path).name


__all__ = [
    "SYNC_LOGGER",
    "SUPPORTED_PROTOCOLS",
    "DEFAULT_PROTOCOL",
    "get_default_netloc",
    "render_remote",
    "remote_file",
    "path_from_uri",
    "pull",
    "push",
]



from pathlib import Path
from typing import Iterable

from enderchest import instance as i

from .utils import instance


class TestInstanceEquality:
    def test_two_instances_with_different_paths_are_not_equal(self):
        assert not i.equals(
            Path(), instance("foo", Path("a")), instance("foo", Path("b"))
        )

    def test_two_instances_with_the_same_path_are_equal(self):
        assert i.equals(Path(), instance("foo", Path("a")), instance("fufu", Path("a")))

    def test_two_instances_with_the_same_abspath_are_equal(self, minecraft_root):
        assert i.equals(
            minecraft_root,
            instance("foo", minecraft_root / "instances" / "axolotl"),
            instance("fufu", Path("instances/axolotl")),
        )

    def test_equality_check_works_with_expanduser(self, minecraft_root, home):
        assert i.equals(
            minecraft_root,
            instance("foo", home / "Library" / "SomeKindOfMC" / "Instances" / "foo"),
            instance("fufu", Path("~/Library/SomeKindOfMC/Instances/foo")),
        )

    def test_equality_check_works_with_minecraft_root_in_home(self, home):
        assert i.equals(
            Path("~/Library/SomeKindOfMC"),
            instance("foo", Path("Instances") / "foo"),
            instance("fufu", Path("~/Library/SomeKindOfMC/Instances/foo")),
        )

    def test_equality_resolves_symlinks(self, home, minecraft_root):
        assert i.equals(
            minecraft_root,
            instance("foo", Path("official_launcher") / ".minecraft"),
            instance("fufu", Path("~/.minecraft")),
        )



"""Specification and configuration of a shulker box"""
import fnmatch
import os
from pathlib import Path
from typing import Any, NamedTuple

import semantic_version as semver

from . import config as cfg
from . import filesystem as fs
from .instance import InstanceSpec, normalize_modloader
from .loggers import CRAFT_LOGGER

_DEFAULT_PRIORITY = 0
_DEFAULT_LINK_DEPTH = 2
_DEFAULT_DO_NOT_LINK = ("shulkerbox.cfg", ".DS_Store")


class ShulkerBox(NamedTuple):
    """Specification of a shulker box

    Parameters
    ----------
    priority : int
        The priority for linking assets in the shulker box (higher priority
        boxes are linked last)
    name : str
        The name of the shulker box (which is incidentally used to break
        priority ties)
    root : Path
        The path to the root of the shulker box
    match_criteria : list-like of tuples
        The parameters for matching instances to this shulker box. Each element
        consists of:

          - the name of the condition
          - the matching values for that condition

        The logic applied is that an instance must match at least one value
        for each condition (so it's ANDing a collection of ORs)
    link_folders : list-like of str
        The folders that should be linked in their entirety
    max_link_depth : int, optional
        By default, non-root-level folders (that is, folders inside of folders)
        will be treated as files for the purpose of linking. Put another way,
        only files with a depth of 2 or less from the shulker root will be
        linked. This behavior can be overridden by explicitly setting
        the `max_link_depth` value, but **this feature is highly experimental**,
        so use it at your own risk.
    do_not_link : list-like of str, optional
        Glob patterns of files that should not be linked. By default, this list
        comprises `shulkerbox.cfg` and `.DS_Store` (for all you mac gamers).

    Notes
    -----
    A shulker box specification is immutable, so making changes (such as
    updating the match criteria) can only be done on copies created via the
    `_replace` method, inherited from the NamedTuple parent class.
    """

    priority: int
    name: str
    root: Path
    match_criteria: tuple[tuple[str, tuple[str, ...]], ...]
    link_folders: tuple[str, ...]
    max_link_depth: int = _DEFAULT_LINK_DEPTH
    do_not_link: tuple[str, ...] = _DEFAULT_DO_NOT_LINK

    @classmethod
    def from_cfg(cls, config_file: Path) -> "ShulkerBox":
        """Parse a shulker box from its config file

        Parameters
        ----------
        config_file : Path
            The path to the config file

        Returns
        -------
        ShulkerBox
            The resulting ShulkerBox

        Raises
        ------
        ValueError
            If the config file at that location cannot be parsed
        FileNotFoundError
            If there is no config file at the specified location
        """
        priority = 0
        max_link_depth = 2
        root = config_file.parent
        name = root.name
        config = cfg.read_cfg(config_file)

        match_criteria: dict[str, tuple[str, ...]] = {}

        for section in config.sections():
            normalized = (
                section.lower().replace(" ", "").replace("-", "").replace("_", "")
            )
            if normalized.endswith("s"):
                normalized = normalized[:-1]  # lazy de-pluralization
            if normalized in ("linkfolder", "folder"):
                normalized = "link-folders"
            if normalized in ("donotlink",):
                normalized = "do-not-link"
            if normalized in ("minecraft", "version", "minecraftversion"):
                normalized = "minecraft"
            if normalized in ("modloader", "loader"):
                normalized = "modloader"
            if normalized in ("instance", "tag", "host"):
                normalized += "s"  # lazy re-pluralization

            if normalized == "propertie":  # lulz
                # TODO check to make sure properties hasn't been read before
                # most of this section gets ignored
                priority = config[section].getint("priority", _DEFAULT_PRIORITY)
                max_link_depth = config[section].getint(
                    "max-link-depth", _DEFAULT_LINK_DEPTH
                )
                # TODO: support specifying filters (and link-folders) in the properties section
                continue
            if normalized in match_criteria:
                raise ValueError(f"{config_file} specifies {normalized} more than once")

            if normalized == "minecraft":
                minecraft_versions = []
                for key, value in config[section].items():
                    if value is None:
                        minecraft_versions.append(key)
                    elif key.lower().strip().startswith("version"):
                        minecraft_versions.append(value)
                    else:  # what happens if you specify ">=1.19" or "=1.12"
                        minecraft_versions.append("=".join((key, value)))
                match_criteria[normalized] = tuple(minecraft_versions)
            elif normalized == "modloader":
                modloaders: set[str] = set()
                for loader in config[section].keys():
                    modloaders.update(normalize_modloader(loader))
                match_criteria[normalized] = tuple(sorted(modloaders))
            else:
                # really hoping delimiter shenanigans doesn't show up anywhere else
                match_criteria[normalized] = tuple(config[section].keys())

        link_folders = match_criteria.pop("link-folders", ())
        do_not_link = match_criteria.pop("do-not-link", _DEFAULT_DO_NOT_LINK)

        return cls(
            priority,
            name,
            root,
            tuple(match_criteria.items()),
            link_folders,
            max_link_depth=max_link_depth,
            do_not_link=do_not_link,
        )

    def write_to_cfg(self, config_file: Path | None = None) -> str:
        """Write this box's configuration to INI

        Parameters
        ----------
        config_file : Path, optional
            The path to the config file, assuming you'd like to write the
            contents to file

        Returns
        -------
        str
            An INI-syntax rendering of this shulker box's config

        Notes
        -----
        The "root" attribute is ignored for this method
        """
        properties: dict[str, Any] = {"priority": self.priority}
        if self.max_link_depth != _DEFAULT_LINK_DEPTH:
            properties["max-link-depth"] = self.max_link_depth

        config = cfg.dumps(
            os.path.join(self.name, fs.SHULKER_BOX_CONFIG_NAME),
            properties,
            **dict(self.match_criteria),
            link_folders=self.link_folders,
            do_not_link=self.do_not_link,
        )

        if config_file:
            config_file.write_text(config)
        return config

    def matches(self, instance: InstanceSpec) -> bool:
        """Determine whether the shulker box matches the given instance

        Parameters
        ----------
        instance : InstanceSpec
            The instance's specification

        Returns
        -------
        bool
            True if the instance matches the shulker box's conditions, False
            otherwise.
        """
        for condition, values in self.match_criteria:
            match condition:  # these should have been normalized on read-in
                case "instances":
                    for value in values:
                        if fnmatch.fnmatchcase(instance.name, value):
                            break
                    else:
                        return False
                case "tags":
                    for value in values:
                        if fnmatch.filter(
                            [tag.lower() for tag in instance.tags], value.lower()
                        ):
                            break
                    else:
                        return False
                case "modloader":
                    for value in values:
                        if fnmatch.fnmatchcase(
                            instance.modloader.lower(),
                            value.lower(),
                        ):
                            break
                    else:
                        return False
                case "minecraft":
                    for value in values:
                        if any(
                            (
                                _matches_version(value, version)
                                for version in instance.minecraft_versions
                            )
                        ):
                            break
                    else:
                        return False
                case "hosts":
                    # this is handled at a higher level
                    pass
                case _:
                    raise NotImplementedError(
                        f"Don't know how to apply match condition {condition}."
                    )
        return True

    def matches_host(self, hostname: str):
        """Determine whether the shulker box should be linked to from the
        current host machine

        Returns
        -------
        bool
            True if the shulker box's hosts spec matches the host, False otherwise.
        """
        for condition, values in self.match_criteria:
            if condition == "hosts":
                if not any(
                    fnmatch.fnmatchcase(hostname.lower(), host_spec.lower())
                    for host_spec in values
                ):
                    return False
        return True


def _matches_version(version_spec: str, version_string: str) -> bool:
    """Determine whether a version spec matches a version string, taking into
    account that neither users nor Mojang rigidly follow semver (or at least
    PEP440)

    Parameters
    ----------
    version_spec : str
        A version specification provided by a user
    version_string : str
        A version string, likely parsed from an instance's configuration

    Returns
    -------
    bool
        True if the spec matches the version, False otherwise

    Notes
    -----
    This method *does not* match snapshots to their corresponding version
    range--for that you're just going to have to be explicit.
    """
    try:
        return semver.SimpleSpec(version_spec).match(semver.Version(version_string))
    except ValueError:
        # fall back to simple fnmatching
        return fnmatch.fnmatchcase(version_string.lower(), version_spec.lower())


DEFAULT_SHULKER_FOLDERS = (  # TODO: customize in enderchest.cfg
    "config",
    "mods",
    "resourcepacks",
    "saves",
    "shaderpacks",
)

STANDARD_LINK_FOLDERS = (  # TODO: customize in enderchest.cfg
    "backups",
    "cachedImages",
    "crash-reports",
    "logs",
    "replay_recordings",
    "screenshots",
    ".bobby",
)


def create_shulker_box(minecraft_root: Path, shulker_box: ShulkerBox) -> None:
    """Create a shulker box folder based on the provided configuration

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder)
    shulker_box : ShulkerBox
        The spec of the box to create

    Notes
    -----
    - The "root" attribute of the ShulkerBox config will be ignored--instead
      the shulker box will be created at
      <minecraft_root>/EnderChest/<shulker box name>
    - This method will fail if there is no EnderChest set up in the minecraft
      root
    - This method does not check to see if there is already a shulker box
      set up at the specificed location--if one exists, its config will
      be overwritten
    """
    root = fs.shulker_box_root(minecraft_root, shulker_box.name)
    root.mkdir(exist_ok=True)

    for folder in (*DEFAULT_SHULKER_FOLDERS, *shulker_box.link_folders):
        CRAFT_LOGGER.debug(f"Creating {root / folder}")
        (root / folder).mkdir(exist_ok=True, parents=True)

    config_path = fs.shulker_box_config(minecraft_root, shulker_box.name)
    shulker_box.write_to_cfg(config_path)
    CRAFT_LOGGER.info(f"Shulker box configuration written to {config_path}")



"""Higher-level functionality around synchronizing with different EnderCherts"""
import logging
from pathlib import Path
from time import sleep
from urllib.parse import ParseResult, urlparse

from . import filesystem as fs
from . import gather
from .enderchest import EnderChest
from .loggers import IMPORTANT, SYNC_LOGGER
from .prompt import confirm
from .sync import path_from_uri, pull, push, remote_file, render_remote


def load_remote_ender_chest(uri: str | ParseResult) -> EnderChest:
    """Load an EnderChest configuration from another machine

    Parameters
    ----------
    uri : URI
        The URI to the remote Minecraft root

    Returns
    -------
    EnderChest
        The remote EnderChest configuration

    Raises
    ------
    ValueError
        If the provided URI is invalid
    RuntimeError
        If the config from the remote EnderChest could not be parsed
    """
    try:
        uri = uri if isinstance(uri, ParseResult) else urlparse(uri)
    except AttributeError as bad_uri:
        raise ValueError(f"{uri} is not a valid URI") from bad_uri

    remote_root = Path(uri.path)
    remote_config_path = fs.ender_chest_config(remote_root, check_exists=False)
    uri = uri._replace(path=remote_config_path.as_posix())

    with remote_file(uri) as remote_config:
        try:
            return EnderChest.from_cfg(remote_config)
        except ValueError as bad_chest:
            raise RuntimeError(
                "The remote EnderChest config downloaded"
                f"from {uri.geturl()} could not be parsed."
            ) from bad_chest


def fetch_remotes_from_a_remote_ender_chest(
    uri: str | ParseResult,
) -> list[tuple[ParseResult, str]]:
    """Grab the list of EnderChests registered with the specified remote EnderChest

    Parameters
    ----------
    uri : URI
        The URI to the remote Minecraft root

    Returns
    -------
    list of (URI, str) tuples
        The URIs of the remote EnderChests, paired with their aliases

    Raises
    ------
    RuntimeError
        If the remote list could not be pulled
    """
    remote_chest = load_remote_ender_chest(uri)
    remotes: list[tuple[ParseResult, str]] = [
        (urlparse(uri) if isinstance(uri, str) else uri, remote_chest.name)
    ]

    remotes.extend(remote_chest.remotes)
    SYNC_LOGGER.info(
        "Loaded the following remotes:\n %s",
        "\n".join(f"  - {render_remote(alias, uri)}" for uri, alias in remotes),
    )

    if len(set(alias for _, alias in remotes)) != len(remotes):
        raise RuntimeError(
            f"There are duplicates aliases in the list of remotes pulled from {uri}"
        )
    return remotes


def sync_with_remotes(
    minecraft_root: Path,
    pull_or_push: str,
    dry_run: bool = False,
    sync_confirm_wait: bool | int | None = None,
    **sync_kwargs,
) -> None:
    """Pull changes from or push changes to remote EnderChests

    Parameters
    ----------
    minecraft_root : Path
        The root directory that your minecraft stuff (or, at least, the one
        that's the parent of your EnderChest folder). This will be used to
        construct relative paths.
    pull_or_push : str
        "pull" or "push"
    dry_run: bool, optional
         Perform a dry run of the sync operation, reporting the operations\
         that will be performed but not actually carrying them out
    sync_confirm_wait : bool or int, optional
        The default behavior when syncing EnderChests is to first perform a dry
        run of every sync operation and then wait 5 seconds before proceeding with the
        real sync. The idea is to give the user time to interrupt the sync if
        the dry run looks wrong. This can be changed by either raising or lowering
        the value of confirm, by disabling the dry-run-first behavior entirely
        (`confirm=False`) or by requiring that the user explicitly confirms
        the sync (`confirm=True`). This default behavior can also be modified
        in the EnderChest config. This parameter will be ignored when performing
        a dry run.
    sync_kwargs
        Any additional arguments that should be passed into the syncing
        operation

    Notes
    -----
    - When pulling changes, this method will try each remote in the order they
      are configured and stop once it has successfully pulled from a remote.

    This method will attempt to push local changes to *every* remote
    """
    if pull_or_push not in ("pull", "push"):
        raise ValueError(
            'Invalid choice for sync operation. Choices are "pull" and "push"'
        )
    try:
        if sync_confirm_wait is None:
            sync_confirm_wait = gather.load_ender_chest(
                minecraft_root
            ).sync_confirm_wait
        this_chest = gather.load_ender_chest(minecraft_root)

        # I know this is redundant, but we want those logs
        remotes = gather.load_ender_chest_remotes(
            minecraft_root, log_level=logging.DEBUG
        )
    except (FileNotFoundError, ValueError) as bad_chest:
        SYNC_LOGGER.error(
            f"Could not load EnderChest from {minecraft_root}:\n  {bad_chest}"
        )
        return
    if not remotes:
        SYNC_LOGGER.error("Enderchest has no remotes. Aborting")
        return  # kinda unnecessary

    synced_somewhere = False
    for remote_uri, alias in remotes:
        if dry_run:
            runs: tuple[bool, ...] = (True,)
        elif sync_confirm_wait is False or sync_confirm_wait <= 0:
            runs = (False,)
        else:
            runs = (True, False)
        for do_dry_run in runs:
            if dry_run:
                prefix = "Simulating an attempt"
            else:
                prefix = "Attempting"
            try:
                if pull_or_push == "pull":
                    SYNC_LOGGER.log(
                        IMPORTANT,
                        f"{prefix} to pull changes from %s",
                        render_remote(alias, remote_uri),
                    )
                    remote_chest = remote_uri._replace(
                        path=urlparse(
                            (
                                fs.ender_chest_folder(
                                    path_from_uri(remote_uri),
                                    check_exists=False,
                                )
                            ).as_uri()
                        ).path
                    )
                    pull(
                        remote_chest,
                        minecraft_root,
                        exclude=[
                            *this_chest.do_not_sync,
                            *(sync_kwargs.pop("exclude", None) or ()),
                        ],
                        dry_run=do_dry_run,
                        **sync_kwargs,
                    )
                else:
                    SYNC_LOGGER.log(
                        IMPORTANT,
                        f"{prefix} to push changes"
                        f" to {render_remote(alias, remote_uri)}",
                    )
                    local_chest = fs.ender_chest_folder(minecraft_root)
                    push(
                        local_chest,
                        remote_uri,
                        exclude=[
                            *this_chest.do_not_sync,
                            *(sync_kwargs.pop("exclude", None) or ()),
                        ],
                        dry_run=do_dry_run,
                        **sync_kwargs,
                    )
            except (
                FileNotFoundError,
                ValueError,
                NotImplementedError,
                TimeoutError,
                RuntimeError,
            ) as sync_fail:
                SYNC_LOGGER.warning(
                    f"Could not sync changes with {render_remote(alias, remote_uri)}:"
                    f"\n  {sync_fail}"
                )
                break
            if do_dry_run == runs[-1]:
                continue
            if sync_confirm_wait is True:
                if not confirm(default=True):
                    SYNC_LOGGER.error("Aborting")
                    return
            else:
                SYNC_LOGGER.debug(f"Waiting for {sync_confirm_wait} seconds")
                sleep(sync_confirm_wait)
        else:
            synced_somewhere = True
            if pull_or_push == "pull":
                break
    if not synced_somewhere:
        SYNC_LOGGER.error("Could not sync with any remote EnderChests")



"""Testing utilities"""
import json
import shutil
from importlib.resources import as_file
from pathlib import Path
from typing import Callable, Iterable

import pytest

from enderchest import EnderChest, InstanceSpec, ShulkerBox
from enderchest import filesystem as fs
from enderchest.instance import normalize_modloader

from . import testing_files

# folders common to both official and MMC-derived .minecraft folders
COMMON_FOLDERS: tuple[str, ...] = (
    "backups",
    "crash_reports",
    "logs",
    "resourcepacks",
    "saves",
    "screenshots",
)

# folders found in the official launcher's .minecraft folder
OFFICIAL_FOLDERS: tuple[str, ...] = (
    *COMMON_FOLDERS,
    "assets",
    "bin",
    "versions",
)

# folders commonly found in the .minecraft folder of MMC-derived instances
MMC_FOLDERS: tuple[str, ...] = (
    *COMMON_FOLDERS,
    "config",
    "coremods",
    "data",
    "patchouli-books",
    "replay_recordings",
    "server-resource-packs",
    "shaderpacks",
    "texturepacks",
    ".bobby",
)

with as_file(testing_files.ENDERCHEST_CONFIG) as enderchest_cfg:
    TESTING_INSTANCES: tuple[InstanceSpec, ...] = tuple(
        EnderChest.from_cfg(enderchest_cfg).instances
    )


def create_mmc_pack_file(
    instance_folder: Path, minecraft_version: str, loader: str | None
) -> None:
    """Generate an MMC pack json and write it to file

    Parameters
    ----------
    instance_folder : Path
        The folder where the pack file should be generated (the parent of
        the .minecraft folder)
    minecraft_version : str
        The minecraft version
    loader : str | None
        The loader (None if it's a vanilla instance)
    """
    if minecraft_version.endswith(".0"):
        # learned the hard way that mmc-pack files don't have use ".0"
        minecraft_version = minecraft_version[:-2]

    components: list[dict[str, str]] = [
        {
            "cachedName": "Minecraft",
            "cachedRequires": "dontreadme",
            "cachedVersion": "dontreadme",
            "important": "dontreadme",
            "uid": "net.minecraft",
            "version": minecraft_version,
        },
    ]

    if loader is not None:
        match loader:
            case "Forge":
                loader_display_name = "Forge"
                loader_uid = "net.minecraftforge"
            case "Fabric":
                loader_display_name = "Fabric Loader"
                loader_uid = "net.fabricmc.fabric-loader"
            case "Quilt":
                loader_display_name = "Quilt"
                loader_uid = "org.quiltmc.quilt-loader"
            case _:
                loader_display_name = loader
                loader_uid = "xxx." + loader.lower().replace(" ", "-")

        components.append(
            {
                "cachedName": loader_display_name,
                "uid": loader_uid,
            }
        )

    instance_folder.mkdir(parents=True, exist_ok=True)

    with (instance_folder / "mmc-pack.json").open("w") as pack_file:
        json.dump(
            {"components": components, "formatVersion": 1},
            pack_file,
            indent=4,
        )


def create_instance_cfg(instance_folder: Path, name: str) -> None:
    """Generate a MultiMC instance config and write it to file

    Parameters
    ----------
    instance_folder : Path
        The folder where the pack file should be generated (the parent of
        the .minecraft folder)
    name : str
        The name of the instance
    """
    instance_folder.mkdir(parents=True, exist_ok=True)

    (instance_folder / "instance.cfg").write_text(
        f"""InstanceType=SevenEightNine
boring=blahblah
iconKey=iconlock
lastLaunchTime=1673878714535
lastTimePlayed=159
name={name}
notes=
totalTimePlayed=2552
"""
    )


def _set_up_minecraft_folder(minecraft_folder: Path, official: bool) -> None:
    """Populate a minecraft folder with a bunch of dummy folders. You probably
    want to use `populate_official_minecraft_folder` or `populate_mmc_instance`
    rather than calling this directly.

    Parameters
    ----------
    minecraft_folder : Path
        Path of the .minecraft folder
    official : bool
        To mimic an MMC-style instance folder, pass in `official=False`. To
        mimic an installation from the official launcher instead, pass in
        `official=True`.
    """
    for folder in OFFICIAL_FOLDERS if official else MMC_FOLDERS:
        (minecraft_folder / folder).mkdir(parents=True)


def populate_official_minecraft_folder(minecraft_folder: Path) -> None:
    """Populate the .minecraft folder created by the official launcher

    Parameters
    ----------
    minecraft_folder : Path
        Path of the .minecraft folder
    """
    _set_up_minecraft_folder(minecraft_folder, official=True)
    with as_file(testing_files.LAUNCHER_PROFILES) as launcher_profiles:
        shutil.copy(launcher_profiles, minecraft_folder)
    with as_file(testing_files.VERSION_MANIFEST) as version_manifest:
        (minecraft_folder / "versions").mkdir(exist_ok=True)
        shutil.copy(version_manifest, minecraft_folder / "versions")


def populate_mmc_instance_folder(
    instance_folder: Path,
    minecraft_version: str,
    loader: str | None,
    name: str | None = None,
) -> None:
    """Populate the .minecraft folder created by an MMC-like launcher

    Parameters
    ----------
    instance_folder : Path
        The folder containing the MMC-like instance's files (the parent of
        the .minecraft folder)
    minecraft_version : str
        The minecraft version
    loader : str | None
        The loader (if vanilla this will be None)
    name : str, optional
        The name of the instance (if different from the name of the folder)
    """
    _set_up_minecraft_folder(instance_folder / ".minecraft", official=False)
    create_mmc_pack_file(
        instance_folder, minecraft_version=minecraft_version, loader=loader
    )
    create_instance_cfg(instance_folder, name or instance_folder.name)


def populate_instances_folder(instances_folder: Path) -> None:
    """Populate an MMC-style "instances" folder according to what's already in
    enderchest.cfg

    Parameters
    ----------
    instances_folder : Path
        The path of the instances folder
    """
    instances_folder.mkdir(parents=True)
    for instance_spec in TESTING_INSTANCES:
        if instance_spec.name == "official":
            continue

        populate_mmc_instance_folder(
            instances_folder / instance_spec.root.parent.name,
            instance_spec.minecraft_versions[0],
            instance_spec.modloader,
        )
    with as_file(testing_files.INSTGROUPS) as instgroups:
        shutil.copy(instgroups, instances_folder)


def pre_populate_enderchest(
    enderchest_folder: Path,
    *shulkers: tuple[str, str],
) -> list[ShulkerBox]:
    """Create an EnderChest folder, pre-populated with the testing enderchest.cfg
    folder and the specified shulker boxes

    Parameters
    ----------
    enderchest_folder : Path
        The path of the EnderChest folder
    *shulkers : 2-tuple of str
        Shulker boxes to populate, with tuple members of:
          - name : the folder name of the shulker
          - config : the contents of the config file

    Returns
    -------
    list of ShulkerBox
        A list of parsed shulker boxes corresponding to the ones rendered
        on the system
    """
    enderchest_folder.mkdir(parents=True, exist_ok=True)
    with as_file(testing_files.ENDERCHEST_CONFIG) as config_file:
        shutil.copy(config_file, enderchest_folder)
    shulker_boxes: list[ShulkerBox] = []
    for shulker_name, shulker_config in shulkers:
        (enderchest_folder / shulker_name).mkdir(parents=True, exist_ok=True)
        config_path = enderchest_folder / shulker_name / fs.SHULKER_BOX_CONFIG_NAME
        with config_path.open("w") as config_file:
            config_file.write(shulker_config)
        shulker_boxes.append(ShulkerBox.from_cfg(config_path))
    return shulker_boxes


# some sample shulker box configs to use with the above
GLOBAL_SHULKER = (
    "global",
    """; global/shulkerbox.cfg

; truly global--could even link to non-minecrafts (one day)
;[minecraft]
;*

[link-folders]
screenshots
backups
crash-reports
logs
""",
)

WILD_UPDATE_SHULKER = (
    "1.19",
    """; 1.19/shulkerbox.cfg
[properties]
name = wild update
priority = 1
notes = Writing it all down

[minecraft]
>=1.19.0,<1.20

[link-folders]
""",
)

VANILLA_SHULKER = (
    "vanilla",
    """; vanilla/shulkerbox.cfg

[properties]
priority = 2
last_modified = 1970-1-1 00:00:00.000000

[minecraft]
*

[modloader]
none
""",
)

OPTIFINE_SHULKER = (
    "optifine",
    """; optifine/shulkerbox.cfg
[properties]
priority = 3

[minecraft]
*

[modloader]
Forge

[link-folders]
shadercache  ; not that I think this is a thing

[do-not-link]
*.local
""",
)

STEAMDECK_SHULKER = (
    "steamdeck",
    """; steamdeck/shulkerbox.cfg
[properties]
priority = 20

[hosts]
steamdeck
""",
)


TESTING_SHULKER_CONFIGS = (
    GLOBAL_SHULKER,
    WILD_UPDATE_SHULKER,
    VANILLA_SHULKER,
    OPTIFINE_SHULKER,
    STEAMDECK_SHULKER,
)

# sometimes you just need a CSV
MATCH_CSV = """, global, 1.19, vanilla, optifine, steamdeck
~              ,   True, True,    True,    False,     False
axolotl        ,   True, False,   True,    False,     False
bee            ,   True, False,  False,     True,     False
chest-boat     ,   True, True,   False,    False,     False
"""


def _parse_match_csv() -> list[tuple[str, str, bool]]:
    """Parse the above table of shulker-to-instance matches"""
    matches: list[tuple[str, str, bool]] = []
    rows = MATCH_CSV.splitlines()
    header = rows.pop(0)
    shulker_boxes = [cell.strip() for cell in header.split(",")[1:]]
    for row in rows:
        cells = [cell.strip() for cell in row.split(",")]
        mc = cells.pop(0)
        for i, cell in enumerate(cells):
            matches.append((shulker_boxes[i], mc, cell == "True"))
    return matches


TESTING_SHULKER_INSTANCE_MATCHES = tuple(_parse_match_csv())


def resolve(path: Path, minecraft_root: Path) -> Path:
    """With all of this file system mocking, path resolution can be a pain
    in the toucans. This method means we should just have to solve for it
    once.

    Parameters
    ----------
    path : Path
        The path you're wanting to check
    minecraft_root : Path
        The minecraft root (from the fixture)

    Returns
    -------
    Path
        The absolute path to the thing you're wanting to check
    """
    if path.expanduser().is_absolute():
        return path.expanduser()
    return (minecraft_root / path).absolute()


def parametrize_over_instances(*instance_names: str):
    """Apply a pytest.mark.parametrize decorator to a test to apply the test
    to each of the specified instances (and apply clear and simple test IDs).

    Parameters
    ----------
    *instance_names : str
        The instances to test. If no instance names are provided, then all
        instances will be included

    Notes
    -----
      - The parametrized tests will be ordered as provided
      - The name of the parametrized argument provided to the test will be "instance"
    """
    instance_lookup = {mc.name: mc for mc in TESTING_INSTANCES}
    if len(instance_names) == 0:
        instance_names = tuple(instance_lookup.keys())

    instances = [instance_lookup[name] for name in instance_names]

    return pytest.mark.parametrize("instance", instances, ids=instance_names)


def scripted_prompt(responses: Iterable[str]) -> Callable[..., str]:
    """Create a replacement for the built-in input() method for use in
    monkeypatching that will process inputs given from a script of responses

    Parameters
    ----------
    responses : list-like of str
        The scripted responses

    Returns
    -------
    function
        A drop-in replacement for the built-in input method
    """
    script = iter(responses)

    def read_from_script(prompt: str | None = None) -> str:
        line = next(script)
        print((prompt or "") + line)
        return line

    return read_from_script


def instance(
    name: str,
    root: Path,
    minecraft_versions: Iterable[str] | None = None,
    modloader: str | None = None,
    tags: Iterable[str] | None = None,
) -> InstanceSpec:
    """Shortcut constructor"""
    return InstanceSpec(
        name, root, tuple(minecraft_versions or ()), modloader or "", tuple(tags or ())
    )


def normalize_instance(mc: InstanceSpec) -> InstanceSpec:
    """Normalize the values inside an instance tuple"""
    return mc._replace(
        # this should be fully checked by instance.equals()
        root=mc.root.expanduser().relative_to(mc.root.expanduser().parent.parent),
        modloader=normalize_modloader(mc.modloader)[0],
        minecraft_versions=tuple(sorted(mc.minecraft_versions)),
        tags=tuple(sorted(tag.lower() for tag in mc.tags)),
    )



"""Command-line interface"""
import inspect
import logging
import os
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from typing import Any, Protocol, Sequence

from . import craft, gather, loggers, place, remote
from ._version import get_versions

# mainly because I think I'm gonna forget what names are canonical (it's the first ones)
_create_aliases = ("craft", "create")
_instance_aliases = tuple(
    alias + plural for alias in ("minecraft", "instance") for plural in ("", "s")
)
_shulker_aliases = ("shulker_box", "shulkerbox", "shulker")
_remote_aliases = tuple(
    alias + plural for alias in ("enderchest", "remote") for plural in ("s", "")
)
_list_aliases = ("inventory", "list")


class Action(Protocol):
    """Common protocol for CLI actions"""

    def __call__(self, minecraft_root: Path, /) -> Any:
        ...


def _place(
    minecraft_root: Path,
    errors: str = "prompt",
    keep_broken_links: bool = False,
    keep_stale_links: bool = False,
    keep_level: int = 0,
    stop_at_first_failure: bool = False,
    ignore_errors: bool = False,
    absolute: bool = False,
    relative: bool = False,
) -> None:
    """Wrapper sort through all the various argument groups"""

    if stop_at_first_failure:
        errors = "abort"
    if ignore_errors:  # elif?
        errors = "ignore"
    # else: errors = errors

    if absolute is True:
        # technically we get this for free already
        relative = False

    if keep_level > 0:
        keep_stale_links = True
    if keep_level > 1:
        keep_broken_links = True

    place.place_ender_chest(
        minecraft_root,
        keep_broken_links=keep_broken_links,
        keep_stale_links=keep_stale_links,
        error_handling=errors,
        relative=relative,
    )


def _craft_shulker_box(minecraft_root: Path, name: str | None = None, **kwargs):
    """Wrapper to handle the fact that name is a required argument"""
    assert name  # it's required by the parser, so this should be fine
    craft.craft_shulker_box(minecraft_root, name, **kwargs)


def _list_instance_boxes(
    minecraft_root: Path, instance_name: str | None = None, **kwargs
):
    """Wrapper to handle the fact that name is a required argument"""
    assert instance_name  # it's required by the parser, so this should be fine
    gather.get_shulker_boxes_matching_instance(minecraft_root, instance_name, **kwargs)


def _list_shulker_box(
    minecraft_root: Path, shulker_box_name: str | None = None, **kwargs
):
    """Wrapper to handle the fact that name is a required argument"""
    assert shulker_box_name  # it's required by the parser, so this should be fine
    gather.get_instances_matching_shulker_box(
        minecraft_root, shulker_box_name, **kwargs
    )


def _update_ender_chest(
    minecraft_root: Path,
    official: bool | None = None,
    mmc: bool | None = None,
    **kwargs,
):
    """Wrapper to resolve the official vs. MultiMC flag"""
    if mmc:
        official = False
    gather.update_ender_chest(minecraft_root, official=official, **kwargs)


def _open(minecraft_root: Path, verbosity: int = 0, **kwargs):
    """Router for open verb"""
    remote.sync_with_remotes(minecraft_root, "pull", verbosity=verbosity, **kwargs)


def _close(minecraft_root: Path, verbosity: int = 0, **kwargs):
    """Router for close verb"""
    remote.sync_with_remotes(minecraft_root, "push", verbosity=verbosity, **kwargs)


ACTIONS: tuple[tuple[tuple[str, ...], str, Action], ...] = (
    # action names (first one is canonical), action description, action method
    (
        sum(((verb, verb + " enderchest") for verb in _create_aliases), ()),
        "create and configure a new EnderChest installation",
        craft.craft_ender_chest,
    ),
    (
        tuple(
            f"{verb} {alias}" for verb in _create_aliases for alias in _shulker_aliases
        ),
        "create and configure a new shulker box",
        _craft_shulker_box,
    ),
    (
        ("place",),
        "link (or update the links) from your instances to your EnderChest",
        _place,
    ),
    (
        tuple("gather " + alias for alias in _instance_aliases),
        "register (or update the registry of) a Minecraft installation",
        _update_ender_chest,
    ),
    (
        tuple("gather " + alias for alias in _remote_aliases),
        "register (or update the registry of) a remote EnderChest",
        _update_ender_chest,
    ),
    (
        # I freely admit this is ridiculous
        sum(
            (
                (
                    verb,
                    *(
                        f"{verb} {alias}"
                        # pluralization is hard
                        for alias in ("shulker_boxes", "shulkerboxes", "shulkers")
                    ),
                )
                for verb in _list_aliases
            ),
            (),
        ),
        "list the shulker boxes inside your Enderchest",
        gather.load_shulker_boxes,
    ),
    (
        tuple(
            f"{verb} {alias}"
            for verb in _list_aliases
            for alias in _instance_aliases
            if alias.endswith("s")
        ),
        "list the minecraft instances registered with your Enderchest",
        gather.load_ender_chest_instances,
    ),
    (
        tuple(
            f"{verb} {alias}"
            for verb in _list_aliases
            for alias in _instance_aliases
            if not alias.endswith("s")
        ),
        "list the shulker boxes that the specified instance links to",
        _list_instance_boxes,
    ),
    (
        tuple(
            f"{verb} {alias}" for verb in _list_aliases for alias in _shulker_aliases
        ),
        "list the minecraft instances that match the specified shulker box",
        _list_shulker_box,
    ),
    (
        tuple(f"{verb} {alias}" for verb in _list_aliases for alias in _remote_aliases),
        "list the other EnderChest installations registered with this EnderChest",
        gather.load_ender_chest_remotes,
    ),
    (
        ("open",),
        "pull changes from other EnderChests",
        _open,
    ),
    (
        ("close",),
        "push changes to other EnderChests",
        _close,
    ),
)


def generate_parsers() -> tuple[ArgumentParser, dict[str, ArgumentParser]]:
    """Generate the command-line parsers

    Returns
    -------
    enderchest_parser : ArgumentParser
        The top-level argument parser responsible for routing arguments to
        specific action parsers
    action_parsers : dict of str to ArgumentParser
        The verb-specific argument parsers
    """
    descriptions: dict[str, str] = {}
    root_description: str = ""
    for commands, description, _ in ACTIONS:
        descriptions[commands[0]] = description
        root_description += f"\n\t{commands[0]}\n\t\tto {description}"

    enderchest_parser = ArgumentParser(
        prog="enderchest",
        description=(
            f"v{get_versions()['version']}\n"
            "\nsyncing and linking for all your Minecraft instances"
        ),
        formatter_class=RawTextHelpFormatter,
    )

    enderchest_parser.add_argument(
        "-v",  # don't worry--this doesn't actually conflict with --verbose
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s v{get_versions()['version']}",
    )

    # these are really just for the sake of --help
    # (the parsed args aren't actually used)
    enderchest_parser.add_argument(
        "action",
        help=f"the action to perform. Options are:{root_description}",
        type=str,
    )
    enderchest_parser.add_argument(
        "arguments",
        nargs="*",
        help="any additional arguments for the specific action."
        " To learn more, try: enderchest {action} -h",
    )

    action_parsers: dict[str, ArgumentParser] = {}
    for verb, description in descriptions.items():
        parser = ArgumentParser(
            prog=f"enderchest {verb}",
            description=description,
        )
        root = parser.add_mutually_exclusive_group()
        root.add_argument(
            "root",
            nargs="?",
            help=(
                "optionally specify your root minecraft directory."
                "  If no path is given, the current working directory will be used."
            ),
            type=Path,
        )
        root.add_argument(
            "--root",
            dest="root_flag",
            help="specify your root minecraft directory",
            type=Path,
        )

        # I'm actually okay with -vvqvqqv hilarity
        parser.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="increase the amount of information that's printed",
        )
        parser.add_argument(
            "--quiet",
            "-q",
            action="count",
            default=0,
            help="decrease the amount of information that's printed",
        )
        action_parsers[verb] = parser

    # craft options
    craft_parser = action_parsers[_create_aliases[0]]
    craft_parser.add_argument(
        "--from",
        dest="copy_from",
        help=(
            "provide the URI (e.g. rsync://deck@my-steam-deck/home/deck/) of a"
            " remote EnderChest installation that can be used"
            " to boostrap the creation of this one."
        ),
    )
    craft_parser.add_argument(
        "-r",
        "--remote",
        dest="remotes",
        action="append",
        help=(
            "provide the URI (e.g. rsync://deck@my-steam-deck/home/deck/) of a"
            " remote EnderChest installation to register with this one"
        ),
    )
    craft_parser.add_argument(
        "-i",
        "--instance",
        dest="instance_search_paths",
        action="append",
        type=Path,
        help="specify a folder to search for Minecraft installations in",
    )
    craft_parser.add_argument(
        "--overwrite",
        action="store_true",
        help=(
            "if there's already an EnderChest installation in this location,"
            " overwrite its configuration"
        ),
    )

    # shulker box craft options
    shulker_craft_parser = action_parsers[f"{_create_aliases[0]} {_shulker_aliases[0]}"]
    shulker_craft_parser.add_argument(
        "name",
        help="specify the name for this shulker box",
    )
    shulker_craft_parser.add_argument(
        "--priority",
        "-p",
        help="specify the link priority for this shulker box (higher = linked later)",
    )
    shulker_craft_parser.add_argument(
        "-i",
        "--instance",
        dest="instances",
        action="append",
        help="only link instances with one of the provided names to this shulker box",
    )
    shulker_craft_parser.add_argument(
        "-t",
        "--tag",
        dest="tags",
        action="append",
        help="only link instances with one of the provided tags to this shulker box",
    )
    shulker_craft_parser.add_argument(
        "-e",
        "--enderchest",
        dest="hosts",
        action="append",
        help=(
            "only link instances registered to one of the provided EnderChest"
            " installations with this shulker box"
        ),
    )
    shulker_craft_parser.add_argument(
        "-l",
        "--folder",
        dest="link_folders",
        action="append",
        help=(
            "specify the name of a folder inside this shulker box"
            " that should be linked completely"
        ),
    )
    shulker_craft_parser.add_argument(
        "--overwrite",
        action="store_true",
        help=(
            "if there's already a shulker box with the specified name,"
            " overwrite its configuration"
        ),
    )

    # place options
    place_parser = action_parsers["place"]
    cleanup = place_parser.add_argument_group()
    cleanup.add_argument(
        "--keep-broken-links",
        action="store_true",
        help="do not remove broken links from instances",
    )
    cleanup.add_argument(
        "--keep-stale-links",
        action="store_true",
        help=(
            "do not remove existing links into the EnderChest,"
            " even if the shulker box or instance spec has changed"
        ),
    )
    cleanup.add_argument(
        "-k",
        dest="keep_level",
        action="count",
        default=0,
        help=(
            "Shorthand for the above cleanup options:"
            " -k will --keep-stale-links,"
            " and -kk will --keep-broken-links as well"
        ),
    )
    error_handling = place_parser.add_mutually_exclusive_group()
    error_handling.add_argument(
        "--stop-at-first-failure",
        "-x",
        action="store_true",
        help="stop linking at the first issue",
    )
    error_handling.add_argument(
        "--ignore-errors", action="store_true", help="ignore any linking errors"
    )
    error_handling.add_argument(
        "--errors",
        "-e",
        choices=(
            "prompt",
            "ignore",
            "skip",
            "skip-instance",
            "skip-shulker-box",
            "abort",
        ),
        default="prompt",
        help=(
            "specify how to handle linking errors"
            " (default behavior is to prompt after every error)"
        ),
    )
    link_type = place_parser.add_mutually_exclusive_group()
    link_type.add_argument(
        "--absolute",
        "-a",
        action="store_true",
        help="use absolute paths for all link targets",
    )
    link_type.add_argument(
        "--relative",
        "-r",
        action="store_true",
        help="use relative paths for all link targets",
    )

    # gather instance options
    gather_instance_parser = action_parsers[f"gather {_instance_aliases[0]}"]
    gather_instance_parser.add_argument(
        "search_paths",
        nargs="+",
        action="extend",
        type=Path,
        help="specify a folder or folders to search for Minecraft installations",
    )
    instance_type = gather_instance_parser.add_mutually_exclusive_group()
    instance_type.add_argument(
        "--official",
        "-o",
        action="store_true",
        help="specify that these are instances managed by the official launcher",
    )
    instance_type.add_argument(
        "--mmc",
        "-m",
        action="store_true",
        help="specify that these are MultiMC-like instances",
    )

    # gather remote options
    gather_remote_parser = action_parsers[f"gather {_remote_aliases[0]}"]
    gather_remote_parser.add_argument(
        "remotes",
        nargs="+",
        action="extend",
        help=(
            "provide URIs (e.g. rsync://deck@my-steam-deck/home/deck/) of any"
            " remote EnderChest installation to register with this one."
            "Note: you should not use this method if the alias (name) of the"
            "remote does not match the remote's hostname (in this example,"
            '"my-steam-deck").'
        ),
    )

    # list instances options

    # list shulkers options

    # list instance box options
    list_instance_boxes_parser = action_parsers[
        f"{_list_aliases[0]} {_instance_aliases[0]}"
    ]
    list_instance_boxes_parser.add_argument(
        "instance_name", help="The name of the minecraft instance to query"
    )

    # list shulker options
    list_shulker_parser = action_parsers[f"{_list_aliases[0]} {_shulker_aliases[0]}"]
    list_shulker_parser.add_argument(
        "shulker_box_name", help="The name of the shulker box to query"
    )

    # open / close options
    for action in ("open", "close"):
        sync_parser = action_parsers[action]

        sync_parser.add_argument(
            "--dry-run",
            action="store_true",
            help=(
                "Perform a dry run of the sync operation,"
                " reporting the operations that will be performed"
                " but not actually carrying them out"
            ),
        )
        sync_parser.add_argument(
            "--exclude",
            "-e",
            nargs="+",
            help="Provide any file patterns you would like to skip syncing",
        )
        sync_parser.add_argument(
            "--timeout",
            "-t",
            type=int,
            help=(
                "Set a maximum number of seconds to try to sync to a remote chest"
                " before giving up and going on to the next one"
            ),
        )
        sync_confirm_wait = sync_parser.add_mutually_exclusive_group()
        sync_confirm_wait.add_argument(
            "--wait",
            "-w",
            dest="sync_confirm_wait",
            type=int,
            help=(
                "The default behavior when syncing EnderChests is to first perform a"
                " dry run of every sync operation and then wait 5 seconds before"
                " proceeding with the real sync. The idea is to give you time to"
                " interrupt the sync if the dry run looks wrong. You can raise or"
                " lower that wait time through this flag. You can also modify it"
                " by editing the enderchest.cfg file."
            ),
        )

    return enderchest_parser, action_parsers


def parse_args(argv: Sequence[str]) -> tuple[Action, Path, int, dict[str, Any]]:
    """Parse the provided command-line options to determine the action to perform and
    the arguments to pass to the action

    Parameters
    ----------
    argv : list-like of str (sys.argv)
        The options passed into the command line

    Returns
    -------
    Callable
        The action method that will be called
    str
        The root of the minecraft folder (parent of the EnderChest)
        where the action will be performed
    int
        The verbosity level of the operation (in terms of log levels)
    dict
        Any additional options that will be given to the action method

    """
    actions: dict[str, Action] = {}
    aliases: dict[str, str] = {}
    for commands, _, method in ACTIONS:
        for command in commands:
            aliases[command] = commands[0]
        actions[commands[0]] = method

    enderchest_parser, action_parsers = generate_parsers()

    _ = enderchest_parser.parse_args(argv[1:2])  # check for --help and --version

    for command in sorted(aliases.keys(), key=lambda x: -len(x)):  # longest first
        if " ".join((*argv[1:], "")).startswith(command + " "):
            action_kwargs = vars(
                action_parsers[aliases[command]].parse_args(
                    argv[1 + len(command.split()) :]
                )
            )

            action = actions[aliases[command]]

            root_arg = action_kwargs.pop("root")
            root_flag = action_kwargs.pop("root_flag")

            verbosity = action_kwargs.pop("verbose") - action_kwargs.pop("quiet")

            argspec = inspect.getfullargspec(action)
            if "verbosity" in argspec.args + argspec.kwonlyargs:
                action_kwargs["verbosity"] = verbosity

            log_level = loggers.verbosity_to_log_level(verbosity)

            return (
                action,
                Path(root_arg or root_flag or os.getcwd()),
                log_level,
                action_kwargs,
            )

    enderchest_parser.print_help(sys.stderr)
    sys.exit(1)


def main():
    """CLI Entrypoint"""
    logger = logging.getLogger(__package__)
    cli_handler = logging.StreamHandler()
    cli_handler.setFormatter(loggers.CLIFormatter())
    logger.addHandler(cli_handler)

    action, root, log_level, kwargs = parse_args(sys.argv)

    # TODO: set log levels per logger based on the command
    cli_handler.setLevel(log_level)

    # TODO: when we add log files, set this to minimum log level across all handlers
    logger.setLevel(log_level)

    action(root, **kwargs)



"""Tests for setting up EnderChests and shulker boxes"""
import logging
import os
from pathlib import Path
from typing import Any

import pytest

from enderchest import EnderChest, ShulkerBox, craft
from enderchest import filesystem as fs
from enderchest.enderchest import create_ender_chest
from enderchest.gather import load_ender_chest
from enderchest.shulker_box import DEFAULT_SHULKER_FOLDERS, STANDARD_LINK_FOLDERS

from . import utils


class TestEnderChestCrafting:
    def test_ender_chest_aborts_right_away_if_minecraft_root_doesnt_exist(
        self, monkeypatch, minecraft_root
    ):
        def mock_prompt(root) -> Any:
            raise AssertionError("I was not to be called.")

        def mock_create(root, chest) -> None:
            raise AssertionError("I was not to be called.")

        def mock_gather(*args, **kwargs) -> list[Any]:
            raise AssertionError("I was not to be called.")

        monkeypatch.setattr(craft, "specify_ender_chest_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_ender_chest", mock_create)
        monkeypatch.setattr(craft, "gather_minecraft_instances", mock_gather)
        monkeypatch.setattr(
            craft, "fetch_remotes_from_a_remote_ender_chest", mock_gather
        )

        craft.craft_ender_chest(minecraft_root / "trunk")

    def test_no_kwargs_routes_to_the_interactive_prompter(self, monkeypatch, tmpdir):
        prompt_log: list[Path] = []

        def mock_prompt(root) -> Any:
            prompt_log.append(root)
            return "MockEnderChest"

        create_log: list[tuple[Path, Any]] = []

        def mock_create(root, chest) -> None:
            create_log.append((root, chest))

        monkeypatch.setattr(craft, "specify_ender_chest_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_ender_chest", mock_create)

        craft.craft_ender_chest(tmpdir)

        assert prompt_log == [tmpdir]
        assert create_log == [(tmpdir, "MockEnderChest")]

    @pytest.mark.parametrize(
        "argument, value",
        (
            ("instance_search_paths", ("~",)),
            ("overwrite", True),
            ("remotes", ("earworm://somewhere/beyond/the-sea",)),
        ),
        ids=("instances", "overwrite", "remotes"),
    )
    def test_any_kwarg_avoids_the_interactive_prompter(
        self,
        monkeypatch,
        argument,
        value,
        minecraft_root,
    ):
        def mock_prompt(root) -> Any:
            raise AssertionError("I was not to be called.")

        create_log: list[tuple[Path, Any]] = []

        def mock_create(root, chest) -> None:
            create_log.append((root, chest))

        def mock_gather(*args, **kwargs) -> list:
            return []

        def mock_fetch_remotes(*args, **kwargs) -> list:
            raise AssertionError("I was not to be called.")

        monkeypatch.setattr(craft, "specify_ender_chest_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_ender_chest", mock_create)
        monkeypatch.setattr(craft, "gather_minecraft_instances", mock_gather)
        monkeypatch.setattr(
            craft, "fetch_remotes_from_a_remote_ender_chest", mock_fetch_remotes
        )

        craft.craft_ender_chest(minecraft_root, **{argument: value})
        assert len(create_log) == 1

    def test_failed_fetch_aborts_the_create(self, monkeypatch, minecraft_root):
        def mock_prompt(root) -> Any:
            raise AssertionError("I was not to be called.")

        def mock_create(root, chest) -> None:
            raise AssertionError("I was not to be called.")

        def mock_gather(*args, **kwargs) -> list:
            return []

        def mock_fetch_remotes(*args, **kwargs) -> list:
            raise RuntimeError("Don't feel like it.")

        monkeypatch.setattr(craft, "specify_ender_chest_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_ender_chest", mock_create)
        monkeypatch.setattr(craft, "gather_minecraft_instances", mock_gather)
        monkeypatch.setattr(
            craft, "fetch_remotes_from_a_remote_ender_chest", mock_fetch_remotes
        )

        craft.craft_ender_chest(minecraft_root, copy_from="prayer://unreachable")

    def test_default_behavior_is_to_prevent_overwrite(self, minecraft_root, caplog):
        create_ender_chest(
            minecraft_root,
            EnderChest(
                "openbagtwo@battlestation" + minecraft_root.absolute().as_posix()
            ),
        )

        original_config = fs.ender_chest_config(minecraft_root).read_text()

        craft.craft_ender_chest(
            minecraft_root, instance_search_paths=(minecraft_root / "instances",)
        )
        error_log = "\n".join(
            record.msg for record in caplog.records if record.levelname == "ERROR"
        )
        assert "There is already an EnderChest installed" in error_log

        assert fs.ender_chest_config(minecraft_root).read_text() == original_config

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_craft_chest_from_config(
        self,
        minecraft_root,
        home,
        caplog,
        monkeypatch,
        capsys,
        root_type,
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        # we'll be testing overwriting
        create_ender_chest(
            root,
            EnderChest(
                "sftp://openbagtwo@battlestation" + minecraft_root.absolute().as_posix()
            ),
        )

        never = utils.scripted_prompt(["no"])
        monkeypatch.setattr("builtins.input", never)

        craft.craft_ender_chest(
            minecraft_root,
            instance_search_paths=(minecraft_root / "instances", home),
            remotes=("rsync://deck@steamdeck/home/deck/minecraft",),
            overwrite=True,
        )

        _ = capsys.readouterr()  # suppress outputs

        assert not [record for record in caplog.records if record.levelname == "ERROR"]

        chest = load_ender_chest(minecraft_root)
        assert len(chest.instances) == 4
        assert len(chest.remotes) == 1

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_giving_default_responses_results_in_the_expected_chest(
        self,
        monkeypatch,
        minecraft_root,
        home,
        capsys,
        caplog,
        root_type,
    ):
        if root_type == "absolute":
            root = minecraft_root
            n_responses = 8
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)
            n_responses = 9  # because now cwd != minecraft root

        script_reader = utils.scripted_prompt([""] * n_responses)
        monkeypatch.setattr("builtins.input", script_reader)

        chest = craft.specify_ender_chest_from_prompt(root)

        _ = capsys.readouterr()  # suppress outputs

        assert not [record for record in caplog.records if record.levelname == "ERROR"]

        assert len(chest.instances) == 4
        assert len(chest.remotes) == 0

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_prompter_gives_you_the_chance_to_back_out(
        self, monkeypatch, tmpdir, capsys
    ):
        script_reader = utils.scripted_prompt(
            ["n", "n", "", "", "me@here", "default", "n"]
        )
        monkeypatch.setattr("builtins.input", script_reader)
        os.chdir(tmpdir)

        with pytest.raises(RuntimeError):
            craft.specify_ender_chest_from_prompt(Path(tmpdir))

        _ = capsys.readouterr()  # suppress outputs

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()


class TestShulkerBoxCrafting:
    def test_default_folders_and_link_folders_do_not_overlap(self):
        assert (
            len(
                [
                    folder
                    for folder in STANDARD_LINK_FOLDERS
                    if folder in DEFAULT_SHULKER_FOLDERS
                ]
            )
            == 0
        )

    def test_no_kwargs_routes_to_the_interactive_prompter(self, monkeypatch):
        prompt_log: list[tuple[Path, str]] = []

        def mock_prompt(root, name) -> Any:
            prompt_log.append((root, name))
            return "MockShulkerBox"

        create_log: list[tuple[Path, Any]] = []

        def mock_create(root, box) -> None:
            create_log.append((root, box))

        monkeypatch.setattr(craft, "specify_shulker_box_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_shulker_box", mock_create)

        craft.craft_shulker_box(Path("/"), "spitty")

        assert prompt_log == [(Path("/"), "spitty")]
        assert create_log == [(Path("/"), "MockShulkerBox")]

    @pytest.mark.parametrize(
        "argument, value",
        (
            ("instances", ("official",)),
            ("overwrite", True),
            ("tags", ("youre", "it")),
            ("hosts", ("elsewhere", "here")),
            ("priority", -12),
        ),
        ids=("instances", "overwrite", "tags", "hosts", "priority"),
    )
    def test_any_kwarg_avoids_the_interactive_prompter(
        self, monkeypatch, argument, value
    ):
        def mock_prompt(root) -> Any:
            raise AssertionError("I was not to be called.")

        create_log: list[tuple[Path, Any]] = []

        def mock_create(root, box) -> None:
            create_log.append((root, box))

        class FakePath:
            def exists(self) -> bool:
                return False

        def mock_fs(*args, **kwargs) -> Any:
            return FakePath()

        monkeypatch.setattr(craft, "specify_shulker_box_from_prompt", mock_prompt)
        monkeypatch.setattr(craft, "create_shulker_box", mock_create)
        monkeypatch.setattr(fs, "shulker_box_config", mock_fs)

        craft.craft_shulker_box(Path("minceraft"), "bacon", **{argument: value})
        assert len(create_log) == 1

    def test_warns_if_you_forgot_to_add_yourself_as_the_host(
        self, monkeypatch, minecraft_root, capsys, caplog
    ):
        utils.pre_populate_enderchest(minecraft_root / "EnderChest")
        script_reader = utils.scripted_prompt(
            [
                "F",
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                "m",
                # while we're here we might as well check setting
                # the other shulker properties
                "cachedImages, logs",
                "-12",
                "somewhere_out_there",
                "",  # default false
                "somewhere_out_there,*",
                "",  # this is the final writeme confirmation
            ]
        )
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft.specify_shulker_box_from_prompt(minecraft_root, "q and a")

        _ = capsys.readouterr()  # suppress outputs

        assert shulker_box == ShulkerBox(
            -12,
            "q and a",
            minecraft_root / "EnderChest" / "q and a",
            match_criteria=(
                ("minecraft", ("*",)),
                ("modloader", ("*",)),
                ("tags", ("*",)),
                ("hosts", ("somewhere_out_there", "*")),
            ),
            link_folders=("cachedImages", "logs"),
        )

        warn_log = "\n".join(
            record.msg for record in caplog.records if record.levelname == "WARNING"
        )

        assert (
            "This shulker box will not link to any instances on this machine"
            in warn_log
        )


class TestPromptByFilter:
    def test_giving_default_responses_results_in_the_expected_shulker_box(
        self, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt([""] * 6)
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_filters(
            ShulkerBox(0, "tester", Path("ignored"), (), ()), utils.TESTING_INSTANCES
        )

        _ = capsys.readouterr()  # suppress outputs

        assert shulker_box.match_criteria == (
            ("minecraft", ("*",)),
            ("modloader", ("*",)),
            ("tags", ("*",)),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_filter_prompt_doesnt_confirm_when_there_are_no_instances(
        self, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt(
            (
                "22w14infinite",
                "N",
                "",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_filters(
            ShulkerBox(0, "tester", Path("ignored"), (), ()), []
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            ("minecraft", ("22w14infinite",)),
            ("modloader", ("",)),
            ("tags", ("*",)),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_filter_prompt_stops_confirming_once_youre_out_of_instances(
        self, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt(
            (
                "22w14infinite",
                "y",
                "b",
                "april-fools*",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_filters(
            ShulkerBox(0, "tester", Path("ignored"), (), ()), utils.TESTING_INSTANCES
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            ("minecraft", ("22w14infinite",)),
            ("modloader", ("Fabric Loader",)),
            ("tags", ("april-fools*",)),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_filter_prompt_lets_you_back_out_of_a_mistake(
        self, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt(
            (
                "1.19.5",
                "",  # default is no
                "1.19",
                "",  # default is yes
                "N,B,Q",
                "",
                "vanilla*",
                "",
            )
        )

        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_filters(
            ShulkerBox(0, "tester", Path("ignored"), (), ()), utils.TESTING_INSTANCES
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            ("minecraft", ("1.19",)),
            ("modloader", ("", "Fabric Loader", "Quilt Loader")),
            ("tags", ("vanilla*",)),
        )

        # check that it was showing the right instances right up until the end
        assert caplog.records[-1].msg % caplog.records[-1].args == (
            """Filters match the instances:
  - official
  - Chest Boat"""
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()


class TestPromptByName:
    def test_confirms_selections(self, monkeypatch, capsys, caplog):
        script_reader = utils.scripted_prompt(
            (
                "abra, kadabra, alakazam",
                "",  # default should be yes here
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_instance_names(
            ShulkerBox(0, "tester", Path("ignored"), (), ())
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            ("instances", ("abra", "kadabra", "alakazam")),
        )

        assert caplog.records[-1].msg % caplog.records[-1].args == (
            """You specified the following instances:
  - abra
  - kadabra
  - alakazam"""
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_default_for_confirmation_is_no_if_empty_response(
        self,
        monkeypatch,
        capsys,
        caplog,
    ):
        script_reader = utils.scripted_prompt(
            (
                "",
                "",
                "",
                "yes",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        shulker_box = craft._prompt_for_instance_names(
            ShulkerBox(0, "tester", Path("ignored"), (), ())
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (("instances", ("*",)),)

        assert (
            "This shulker box will be applied to all instances"
            in caplog.records[-1].msg
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()


class TestPromptByNumber:
    def test_defaults_results_in_explicit_enumeration(self, monkeypatch, capsys):
        script_reader = utils.scripted_prompt(
            (
                "",
                "",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        def get_instances():
            return utils.TESTING_INSTANCES

        shulker_box = craft._prompt_for_instance_numbers(
            ShulkerBox(0, "tester", Path("ignored"), (), ()),
            get_instances(),
            get_instances,
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            ("instances", tuple(instance.name for instance in utils.TESTING_INSTANCES)),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_number_selection_supports_explicit_numbers_and_ranges(
        self, monkeypatch, capsys
    ):
        script_reader = utils.scripted_prompt(
            (
                "4, 2 - 3",
                "",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        def get_instances():
            return utils.TESTING_INSTANCES

        shulker_box = craft._prompt_for_instance_numbers(
            ShulkerBox(0, "tester", Path("ignored"), (), ()),
            get_instances(),
            get_instances,
        )

        _ = capsys.readouterr()

        assert shulker_box.match_criteria == (
            (
                "instances",
                tuple(instance.name for instance in utils.TESTING_INSTANCES[1:4]),
            ),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()

    def test_invalid_selection_reminds_you_of_the_instance_list(
        self, monkeypatch, capsys, caplog
    ):
        script_reader = utils.scripted_prompt(
            (
                "7",
                "1, 3",
                "",
            )
        )
        monkeypatch.setattr("builtins.input", script_reader)

        def get_instances():
            logging.info(
                """These are the instances that are currently registered:
  1. official (~/.minecraft)
  2. axolotl (instances/axolotl/.minecraft)
  3. bee (instances/bee/.minecraft)
  4. Chest Boat (instances/chest-boat/.minecraft)"""
            )
            return utils.TESTING_INSTANCES

        shulker_box = craft._prompt_for_instance_numbers(
            ShulkerBox(0, "tester", Path("ignored"), (), ()),
            get_instances(),
            get_instances,
        )

        _ = capsys.readouterr()

        assert """Invalid selection: 7 is out of range

These are the instances that are currently registered:
  1. official (~/.minecraft)
  2. axolotl (instances/axolotl/.minecraft)
  3. bee (instances/bee/.minecraft)
  4. Chest Boat (instances/chest-boat/.minecraft)""" in "\n".join(
            record.msg for record in caplog.records
        )

        assert shulker_box.match_criteria == (
            (
                "instances",
                ("official", "bee"),
            ),
        )

        # make sure all responses were used
        with pytest.raises(StopIteration):
            script_reader()



"""Tests around file transfer functionality."""
import logging
import os
import shutil
from pathlib import Path
from urllib.parse import urlparse

import pytest

from enderchest import craft
from enderchest import filesystem as fs
from enderchest import gather
from enderchest import remote as r
from enderchest.sync import file, path_from_uri

from . import utils


class TestPathFromURI:
    def test_roundtrip(self, tmpdir):
        original_path = Path(tmpdir) / "this has a space in it" / "(="
        assert path_from_uri(urlparse(original_path.as_uri())) == original_path


@pytest.mark.xfail(
    not shutil.which("rsync"), reason="rsync is not installed on this system"
)  # TODO: remove xfail once this method has been moved to a different module
class TestURIToSSH:
    @pytest.fixture(scope="class")
    def rsync_module(self):
        from enderchest.sync import rsync

        yield rsync

    def test_simple_parse(self, rsync_module):
        address = rsync_module.uri_to_ssh(
            urlparse("rsync://openbagtwo@couchgaming:22/home/openbagtwo/minecraft")
        )
        assert address == "openbagtwo@couchgaming:22:/home/openbagtwo/minecraft"

    def test_no_username_parse(self, rsync_module):
        address = rsync_module.uri_to_ssh(
            urlparse("rsync://steamdeck/home/openbagtwo/minecraft")
        )
        assert address == "steamdeck:/home/openbagtwo/minecraft"

    def test_no_netloc_parse(self, rsync_module):
        address = rsync_module.uri_to_ssh(
            urlparse("rsync:///mnt/external/minecraft-bkp")
        )
        assert address == "localhost:/mnt/external/minecraft-bkp"

    def test_no_hostname_parse(self, rsync_module):
        """Can't believe this is a valid URI"""
        address = rsync_module.uri_to_ssh(urlparse("rsync://nugget@/home/nugget/"))
        assert address == "nugget@localhost:/home/nugget"


class TestFileIgnorePatternBuilder:
    def test_simple_match(self):
        assert file.ignore_patterns("hello")(
            "greetings", ("bonjour", "hello", "hellooooo")
        ) == {"hello"}

    def test_wildcard(self):
        assert file.ignore_patterns("hel*")(
            "responses", ("como sa va", "hello", "hell no", "help")
        ) == {"hello", "hell no", "help"}

    def test_multi_pattern_match(self):
        assert file.ignore_patterns("hel*", "bye")(
            "responses", ("hello", "goodbye", "hellooo", "bye")
        ) == {"hello", "hellooo", "bye"}

    def test_full_path_check(self):
        ignore = file.ignore_patterns(os.path.join("root", "branch"))
        assert (
            ignore("root", ("branch", "trunk")),
            ignore("trunk", ("branch", "leaf")),
        ) == ({"branch"}, set())

    def test_match_is_performed_on_the_end(self):
        ignore = file.ignore_patterns(os.path.join("root", "branch"), "leaf")
        assert (
            ignore(os.path.join("tree", "root"), ("branch", "trunk")),
            ignore("wind", ("leaf", "blows")),
        ) == ({"branch"}, {"leaf"})


class TestFileSync:
    protocol = "file"

    @pytest.fixture(autouse=True)
    def check_output_suppression(self, capfd):
        # teardown that helps me figure out where I need to add a -q flag
        yield
        assert capfd.readouterr().out == ""

    @pytest.fixture
    def remote(self, tmp_path, minecraft_root, home):
        utils.pre_populate_enderchest(
            minecraft_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS
        )

        local = gather.load_ender_chest(minecraft_root)
        local.sync_confirm_wait = False
        local.write_to_cfg(fs.ender_chest_config(minecraft_root))

        another_root = tmp_path / "not so remote"

        (another_root / "EnderChest").mkdir(parents=True)

        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[1:]:
            shutil.copytree(
                minecraft_root / "EnderChest" / folder_name,
                another_root / "EnderChest" / folder_name,
                symlinks=True,
            )

        (minecraft_root / "safe_keeping").mkdir()
        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[:-2]:
            shutil.copytree(
                minecraft_root / "EnderChest" / folder_name,
                minecraft_root / "safe_keeping" / folder_name,
                symlinks=True,
            )
        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS[-2:]:
            shutil.move(
                minecraft_root / "EnderChest" / folder_name,
                minecraft_root / "safe_keeping" / folder_name,
            )

        utils.pre_populate_enderchest(
            another_root / "EnderChest", *utils.TESTING_SHULKER_CONFIGS[1:]
        )

        not_so_remote = gather.load_ender_chest(another_root)
        not_so_remote.name = "closer than you think"
        not_so_remote._uri = not_so_remote._uri._replace(scheme=self.protocol)
        not_so_remote.register_remote(local._uri)
        not_so_remote.register_remote("ipoac://yoursoul@birdhouse/minecraft")
        not_so_remote.write_to_cfg(another_root / "EnderChest" / "enderchest.cfg")

        (another_root / "chest monster" / "worlds" / "olam").mkdir(parents=True)
        (another_root / "chest monster" / "worlds" / "olam" / "level.dat").write_text(
            "dootdootdoot"
        )

        (another_root / "EnderChest" / "1.19" / "saves" / "olam").unlink()
        (another_root / "EnderChest" / "1.19" / "saves" / "olam").symlink_to(
            another_root / "chest monster" / "worlds" / "olam", target_is_directory=True
        )

        (another_root / "EnderChest" / "1.19" / ".bobby").mkdir(exist_ok=True)
        (another_root / "EnderChest" / "1.19" / ".bobby" / "chunk").write_text(
            "chunky\n"
        )

        for root in (minecraft_root, another_root):
            (root / "EnderChest" / "vanilla" / "conflict").mkdir()

        (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).write_text("sparkle")
        (
            another_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).write_text("lab-grown!")

        (another_root / "EnderChest" / "optifine" / "mods").mkdir(
            parents=True, exist_ok=True
        )
        (another_root / "EnderChest" / "optifine" / "mods" / "optifine.jar").write_text(
            "it's okay"
        )

        yield not_so_remote._uri

        assert (
            another_root / "chest monster" / "worlds" / "olam" / "level.dat"
        ).read_text() == "dootdootdoot"

        # now put everything local back the way it was supposed to be before
        # the conftest teardown freaks out

        for folder_name, _ in utils.TESTING_SHULKER_CONFIGS:
            shutil.move(
                minecraft_root / "safe_keeping" / folder_name,
                minecraft_root / "EnderChest" / folder_name,
            )

        # and then I apparently still need to redo some symlinks?
        for path, target in (
            (
                minecraft_root / "EnderChest" / "global" / "crash-reports",
                minecraft_root / "crash-reports",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "resourcepacks",
                minecraft_root / "resourcepacks",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "screenshots",
                home / "Pictures" / "Screenshots",
            ),
            (
                minecraft_root / "EnderChest" / "global" / "saves" / "test",
                minecraft_root / "worlds" / "testbench",
            ),
            (
                minecraft_root / "EnderChest" / "1.19" / "saves" / "olam",
                minecraft_root / "worlds" / "olam",
            ),
        ):
            path.unlink(missing_ok=True)
            path.symlink_to(target, target_is_directory=target.is_dir())

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_create_from_remote_chest(
        self, remote, minecraft_root, monkeypatch, root_type, capfd
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        craft.craft_ender_chest(root, copy_from=remote.geturl(), overwrite=True)

        assert gather.load_ender_chest_remotes(minecraft_root) == [
            (remote, "closer than you think"),
            (urlparse("ipoac://yoursoul@birdhouse/minecraft"), "birdhouse"),
        ]

        # test that the enderchest spec-fetch is quiet by default
        assert capfd.readouterr().out == ""

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_open_grabs_files_from_upstream(
        self, minecraft_root, remote, monkeypatch, root_type
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        gather.update_ender_chest(root, remotes=(remote,))
        r.sync_with_remotes(root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).read_text() == "it's okay"
        assert not (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).is_symlink()

    def test_open_dry_run_does_nothing(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        assert not (
            minecraft_root / "EnderChest" / "optifine" / "mods" / "optifine.jar"
        ).exists()

    def test_open_overwrites_local_files(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).read_text() == "lab-grown!"
        assert not (
            minecraft_root / "EnderChest" / "vanilla" / "conflict" / "diamond.png"
        ).is_symlink()

    def test_open_copies_over_symlinks(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "1.19" / "saves" / "olam"
        ).resolve() != (minecraft_root / "worlds" / "olam")

    def test_open_processes_deletions_from_upstream(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert not (minecraft_root / "EnderChest" / "global").exists()

    def test_open_does_not_overwrite_enderchest_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        original_config = fs.ender_chest_config(minecraft_root).read_text()
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert original_config == fs.ender_chest_config(minecraft_root).read_text()

    def test_open_not_touch_top_level_dot_folders_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / ".git" / "log"
        ).read_text() == "i committed some stuff\n"

    def test_open_will_sync_dot_folders_within_a_shulker_box(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        assert (
            minecraft_root / "EnderChest" / "1.19" / ".bobby" / "chunk"
        ).read_text() == "chunky\n"

    @pytest.mark.parametrize("root_type", ("absolute", "relative"))
    def test_close_overwrites_with_changes_from_local(
        self, minecraft_root, remote, monkeypatch, root_type
    ):
        if root_type == "absolute":
            root = minecraft_root
        else:
            root = Path(minecraft_root.name)
            monkeypatch.chdir(minecraft_root.parent)

        gather.update_ender_chest(root, remotes=(remote,))
        r.sync_with_remotes(root, "push", verbosity=-1)
        assert (
            path_from_uri(remote)
            / "EnderChest"
            / "vanilla"
            / "conflict"
            / "diamond.png"
        ).read_text() == "sparkle"

    def test_close_dry_run_does_nothing(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", dry_run=True)
        assert (
            path_from_uri(remote)
            / "EnderChest"
            / "vanilla"
            / "conflict"
            / "diamond.png"
        ).read_text() == "lab-grown!"

    def test_close_deletes_remote_copies_when_locals_are_deleted(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert not (path_from_uri(remote) / "EnderChest" / "optifine").exists()

    def test_close_not_touch_top_level_dot_folders_by_default(
        self, minecraft_root, remote
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert not (path_from_uri(remote) / "EnderChest" / ".git").exists()

    def test_chest_obeys_its_own_ignore_list(self, minecraft_root, remote):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))

        chest = gather.load_ender_chest(minecraft_root)
        chest.do_not_sync = ["EnderChest/enderchest.cfg"]
        chest.write_to_cfg(fs.ender_chest_config(minecraft_root))

        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        assert (
            path_from_uri(remote) / "EnderChest" / ".git" / "log"
        ).read_text() == "i committed some stuff\n"

    def test_open_stops_at_first_successful_sync(self, minecraft_root, remote, caplog):
        gather.update_ender_chest(
            minecraft_root, remotes=(remote, "prayer://unreachable")
        )
        r.sync_with_remotes(minecraft_root, "pull", verbosity=-1)
        warnings = [
            record.msg for record in caplog.records if record.levelname == "WARNING"
        ]

        assert len(warnings) == 0

    def test_close_will_attempt_to_push_to_all(self, minecraft_root, remote, caplog):
        gather.update_ender_chest(
            minecraft_root, remotes=(remote, "prayer://unreachable")
        )
        r.sync_with_remotes(minecraft_root, "push", verbosity=-1)
        warnings = [
            record.msg for record in caplog.records if record.levelname == "WARNING"
        ]

        assert len(warnings) == 1
        assert "Could not sync changes with prayer://unreachable" in warnings[0]


@pytest.mark.xfail(
    not shutil.which("rsync"), reason="rsync is not installed on this system"
)
class TestRsyncSync(TestFileSync):
    protocol = "rsync"

    def test_rsync_summary_summarizes_at_the_shulker_box_level(
        self, minecraft_root, remote, caplog
    ):
        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        info_log = ""
        debug_log = ""
        for record in caplog.records:
            if record.levelname == "INFO":
                info_log += record.msg % record.args + "\n"
            elif record.levelname == "DEBUG":
                debug_log += record.msg % record.args + "\n"

        # meta-test--make sure that the creation is actually happening
        assert (
            "+++ " + os.path.sep.join(("EnderChest", "1.19", ".bobby", "chunk"))
            in debug_log
        )

        # meta-test--make sure nothing in the report at the shulker+1 level
        assert os.path.sep.join(("EnderChest", "1.19", ".bobby")) not in info_log

        assert (
            f"Within EnderChest{os.path.sep}1.19..."
            "\n  - Creating 1 file"
            "\n  - Updating 1 file"  # there's also the save symlink
            "\n  - Deleting 0 files"
        ) in info_log

    def test_rsync_summary_doesnt_report_details_if_the_whole_shulker_is_being_created(
        self, minecraft_root, remote, caplog
    ):
        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, "pull", dry_run=True)
        info_log = ""
        debug_log = ""
        for record in caplog.records:
            if record.levelname == "INFO":
                info_log += record.msg + "\n"
            elif record.levelname == "DEBUG":
                debug_log += record.msg + "\n"

        # meta-test--make sure that the creation is actually happening
        assert (
            "+++ "
            + os.path.sep.join(("EnderChest", "optifine", "mods", "optifine.jar"))
            in debug_log
        )

        # meta-test--make sure nothing in the report at the shulker+1 level
        assert os.path.sep.join(("EnderChest", "optifine", "mods")) not in info_log

        assert f"Creating EnderChest{os.path.sep}optifine" in info_log

    @pytest.mark.parametrize("verbosity", ("v", "vv", "vvv"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_verbose_dry_run_doesnt_summarize(
        self, monkeypatch, minecraft_root, remote, caplog, op, verbosity
    ):
        def mock_summarize(*args, **kwargs):
            raise AssertionError("I was not to be called")

        from enderchest.sync import rsync

        monkeypatch.setattr(rsync, "summarize_rsync_report", mock_summarize)

        caplog.set_level(logging.DEBUG)
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=len(verbosity))

        debug_log = "\n".join(
            record.msg for record in caplog.records if record.levelname == "DEBUG"
        )

        # this wouldn't be in the summary
        assert f"EnderChest{os.sep}global{os.sep}config" in debug_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_quiet_dry_run_still_reports_stats(
        self, minecraft_root, remote, caplog, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=-1)

        printed_log = "\n".join(
            record.msg for record in caplog.records if record.levelno > logging.INFO
        )

        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_super_quiet_dry_run_still_reports_stats(
        self, minecraft_root, remote, caplog, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, dry_run=True, verbosity=-1)

        printed_log = "\n".join(
            record.msg for record in caplog.records if record.levelno > logging.INFO
        )

        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_regular_sync_only_reports_overall_progress(
        self, minecraft_root, remote, capfd, op
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op)

        printed_log = capfd.readouterr().out

        assert f"EnderChest{os.sep}global" not in printed_log

        # but, like, make sure that it prints *something*
        assert "Number of created files" in printed_log

    @pytest.mark.parametrize("verbosity", ("v", "vv", "vvv"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_verbose_sync_reports_file_level_progress(
        self, minecraft_root, remote, capfd, op, verbosity
    ):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, verbosity=len(verbosity))

        printed_log = capfd.readouterr().out

        assert f"EnderChest{os.sep}global" in printed_log
        assert "xfr#2, to-chk=" in printed_log

    @pytest.mark.parametrize("quietude", ("q", "qq", "qqq"))
    @pytest.mark.parametrize("op", ("pull", "push"))
    def test_quiet_sync_is_silent(self, minecraft_root, remote, capfd, op, quietude):
        gather.update_ender_chest(minecraft_root, remotes=(remote,))
        r.sync_with_remotes(minecraft_root, op, verbosity=-len(quietude))

        printed_log = capfd.readouterr().out

        assert printed_log == ""



"""Non-implementation-specific syncing utilities"""
import getpass
import os
import socket
from pathlib import Path
from urllib.parse import ParseResult, unquote
from urllib.request import url2pathname


def get_default_netloc() -> str:
    """Compile a netloc from environment variables, etc.

    Returns
    -------
    str
        The default netloc, which is {user}@{hostname}
    """
    return f"{getpass.getuser()}@{socket.gethostname()}".lower()


def path_from_uri(uri: ParseResult) -> Path:
    """Extract and unquote the path component of a URI to turn it into a pathlib.Path

    h/t https://stackoverflow.com/a/61922504

    Parameters
    ----------
    uri : ParseResult
        The parsed URI to extract the path from

    Returns
    -------
    Path
        The path part of the URI as a Path
    """
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=uri.netloc)
    return Path(os.path.abspath(os.path.join(host, url2pathname(unquote(uri.path)))))


def render_remote(alias: str, uri: ParseResult) -> str:
    """Render a remote to a descriptive string

    Parameters
    ----------
    alias : str
        The name of the remote
    uri : ParseResult
        The parsed URI for the remote

    Returns
    -------
    str
        {uri_string} [({alias})]}
            (if different from the URI hostname)
    """
    uri_string = uri.geturl()

    if uri.hostname != alias:
        uri_string += f" ({alias})"
    return uri_string
